# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Healthy and unhealthy citrus fruits and leaves dataset."""

import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@article{rauf2019citrus,
  title={A citrus fruits and leaves dataset for detection and classification of
citrus diseases through machine learning},
  author={Rauf, Hafiz Tayyab and Saleem, Basharat Ali and Lali, M Ikram Ullah
and Khan, Muhammad Attique and Sharif, Muhammad and Bukhari, Syed Ahmad Chan},
  journal={Data in brief},
  volume={26},
  pages={104340},
  year={2019},
  publisher={Elsevier}
}
"""

_DESCRIPTION = """
The original citrus dataset contains 759 images of healthy and unhealthy citrus
fruits and leaves. However, for now we only export 594 images of citrus leaves
with the following labels: Black Spot, Canker, Greening, and Healthy. The
exported images are in PNG format and have 256x256 pixels.

NOTE: Leaf images with Melanose label were dropped due to very small count and
other non-leaf images being present in the same directory.

Dataset URL: https://data.mendeley.com/datasets/3f83gxmv57/2
License: http://creativecommons.org/licenses/by/4.0
"""

_URL = "https://data.mendeley.com/public-files/datasets/3f83gxmv57/files/53398b67-6f0e-4a67-8384-e2b574b2ebf4/file_downloaded"
_LEAVES_LABELS = ["Black spot", "canker", "greening", "healthy"]


class CitrusLeaves(tfds.core.GeneratorBasedBuilder):
  """Subset of the citrus dataset with just leaves."""

  VERSION = tfds.core.Version("0.1.2")
  RELEASE_NOTES = {
      "0.1.1": "Citrus Leaves dataset",
      "0.1.2": "Website URL update",
  }

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(),
            "image/filename": tfds.features.Text(),
            "label": tfds.features.ClassLabel(names=_LEAVES_LABELS),
        }),
        supervised_keys=("image", "label"),
        homepage="https://data.mendeley.com/datasets/3f83gxmv57/2",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(_URL)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN, gen_kwargs={"datapath": path}
        )
    ]

  def _generate_examples(self, datapath):
    """Yields examples based on the passed split index."""
    for label in _LEAVES_LABELS:
      # The real dataset has spaces in directories (label names), which causes
      # fake data test to fail due objfs not handling whitespace in paths. The
      # solution is to replace spaces with underscores in fake data directories
      # and then not care whether a character is a space or an underscore.
      fuzzy_label = label.replace(" ", "[_ ]")
      glob_path = os.path.join(datapath, "Citrus/Leaves", fuzzy_label, "*.png")
      for fpath in tf.io.gfile.glob(glob_path):
        fname = os.path.basename(fpath)
        record = {
            "image": fpath,
            "image/filename": fname,
            "label": label,
        }
        yield "{}/{}".format(label, fname), record



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""gsm8k dataset."""

import json

from etils import epath
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
A dataset of 8.5K high quality linguistically diverse grade school math word problems.
"""

_CITATION = """
@misc{cobbe2021training,
      title={Training Verifiers to Solve Math Word Problems},
      author={Karl Cobbe and Vineet Kosaraju and Mohammad Bavarian and Jacob Hilton and Reiichiro Nakano and Christopher Hesse and John Schulman},
      year={2021},
      eprint={2110.14168},
      archivePrefix={arXiv},
      primaryClass={cs.LG}
}
"""

_URL_PREFIX = 'https://raw.githubusercontent.com/openai/grade-school-math/master/grade_school_math/data/'
_URLS = {
    k: f'{_URL_PREFIX}{k}.jsonl'
    for k in ['train', 'test', 'train_socratic', 'test_socratic']
}

_FEATURES = ['question', 'answer', 'annotation', 'short_answer']


class Gsm8k(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for gsm8k dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(
            {k: tfds.features.Text() for k in _FEATURES}
        ),
        supervised_keys=None,
        homepage='https://github.com/openai/grade-school-math',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    extracted = dl_manager.download_and_extract(_URLS)
    return {k: self._generate_examples(v) for k, v in extracted.items()}  # pytype: disable=wrong-arg-types  # always-use-return-annotations

  def _generate_examples(self, path: str):
    """Yields examples."""
    with epath.Path(path).open() as f:
      for i, line in enumerate(f):
        ex = json.loads(line)
        ex['annotation'], ex['short_answer'] = ex['answer'].split('#### ')
        yield i, ex



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""diamonds dataset."""

from __future__ import annotations

import collections

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
This classic dataset contains physical attributes and prices of 53940 diamonds.

Attributes:

  * price: Price in US dollars.
  * carat: Weight of the diamond.
  * cut: Cut quality (ordered worst to best).
  * color: Color of the diamond (ordered best to worst).
  * clarity: Clarity of the diamond (ordered worst to best).
  * x: Length in mm.
  * y: Width in mm.
  * z: Depth in mm.
  * depth: Total depth percentage: 100 * z / mean(x, y)
  * table: Width of the top of the diamond relative to the widest point.

"""

_CITATION = """
@Book{,
  author = {Hadley Wickham},
  title = {ggplot2: Elegant Graphics for Data Analysis},
  publisher = {Springer-Verlag New York},
  year = {2016},
  isbn = {978-3-319-24277-4},
  url = {https://ggplot2.tidyverse.org},
}
"""

_CUTS = ('Fair', 'Good', 'Very Good', 'Premium', 'Ideal')
_COLORS = ('D', 'E', 'F', 'G', 'H', 'I', 'J')
_CLARITY = ('I1', 'SI2', 'SI1', 'VS2', 'VS1', 'VVS2', 'VVS1', 'IF')


def _features():
  return collections.OrderedDict((
      ('carat', np.float32),
      ('cut', tfds.features.ClassLabel(names=_CUTS)),
      ('color', tfds.features.ClassLabel(names=_COLORS)),
      ('clarity', tfds.features.ClassLabel(names=_CLARITY)),
      ('x', np.float32),
      ('y', np.float32),
      ('z', np.float32),
      ('depth', np.float32),
      ('table', np.float32),
  ))


_URL = 'https://raw.githubusercontent.com/tidyverse/ggplot2/main/data-raw/diamonds.csv'


class Diamonds(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for diamonds dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'features': {k: v for k, v in _features().items()},
            'price': np.float32,
        }),
        supervised_keys=('features', 'price'),
        homepage='https://ggplot2.tidyverse.org/reference/diamonds.html',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    data = dl_manager.download({'data': _URL})
    # There is no predefined train/val/test split for this dataset.
    return {tfds.Split.TRAIN: self._generate_examples(file_path=data['data'])}

  def _generate_examples(self, file_path):
    """Yields examples."""
    pd = tfds.core.lazy_imports.pandas

    with epath.Path(file_path).open() as f:
      df = pd.read_csv(f)

    for row in df.itertuples():
      yield row.Index, {
          'features': {k: getattr(row, k) for k in _features().keys()},
          'price': row.price,
      }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""dolphin number word dataset."""

from __future__ import annotations

import json

import numpy as np
import tensorflow_datasets.public_api as tfds

_DS_PATH = 'https://www.microsoft.com/en-us/research/uploads/prod/2016/02//dolphin-number_word_std.zip'

_DESCRIPTION = """
Dolphin Math Word Problem dataset (2015), as presented in https://www.microsoft.com/en-us/research/uploads/prod/2016/02//dolphin-sigmadolphin.datasets.pdf
"""

_CITATION = """
@inproceedings{inproceedings,
author = {Shi, Shuming and Wang, Yuehui and Lin, Chin-Yew and Liu, Xiaojiang and Rui, Yong},
year = {2015},
month = {09},
pages = {},
title = {Automatically Solving Number Word Problems by Semantic Parsing and Reasoning},
doi = {10.18653/v1/D15-1135}
}
"""


class DolphinNumberWord(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for dolphin_number_word problem dataset."""

  __count__ = 0

  VERSION = tfds.core.Version('0.0.2')
  RELEASE_NOTES = {
      '0.0.1': 'Initial release.',
      '0.0.2': (
          'RaggedTensor fix. Equations and Sources represented as a single'
          'string with components delimited by spaces'
      ),
      '0.0.3': (
          'Reintroduced logic to handle edge-case involving examples without '
          'sources.'
      ),
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        # Per author's recommendation, we discard the ans_simple field.
        features=tfds.features.FeaturesDict({
            'id': tfds.features.Text(),
            'index': np.int32,
            'text': tfds.features.Text(),
            'sources': tfds.features.Text(),  # Flattened list of str.
            'equations': tfds.features.Text(),  # Flattened list of str.
            'ans': tfds.features.Text(),
        }),
        supervised_keys=('text', 'ans'),  # Alternatively text, ans.
        homepage=(
            'https://www.microsoft.com/en-us/research/project/sigmadolphin-2/'
        ),
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(_DS_PATH)
    return {
        'train': self._generate_examples(path, '.dev'),
        'test': self._generate_examples(path, '.test'),
    }

  def _generate_examples(self, json_paths, split):
    """Parses available JSON files in a given split, returns TF examples."""
    # Currently assume all files in archive are json-formatted.
    for file in json_paths.glob('*' + split + '*'):  # pytype: disable=attribute-error  # gen-stub-imports
      json_string = file.read_text()
      examples = json.loads(json_string)
      for e in examples:
        # Per author's recommendation, we discard the ans_simple field.
        e.pop('ans_simple')
        # 'equations' and 'sources' will be a RaggedTensor unless we flatten
        # them into a single string. RaggedTensors are not compatible with
        # t5 tasks.
        if 'sources' in e:
          e['sources'] = '  '.join(e['sources'])
        else:
          e['sources'] = ''
        e['equations'] = '  '.join(e['equations'])
        # Dataset appears to have duplicate entries, we add an internal
        # count to the provided entry's ID to bypass this error.
        e['id'] = e['id'] + str(self.__count__)
        self.__count__ += 1
        yield e['id'], e



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.features.dataset_feature."""

from absl.testing import parameterized
import numpy as np
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import decode as decode_lib
from tensorflow_datasets.core import features as feature_lib


class IncrementDecoder(decode_lib.Decoder):
  """Basic decoder that just adds 1 to the encoded example."""

  def decode_example(self, serialized_example):
    return serialized_example + 1


class DatasetDictFeatureTest(
    parameterized.TestCase, testing.FeatureExpectationsTestCase
):

  @parameterized.parameters((np.int32), (tf.int32))
  def test_int(self, dtype):
    self.assertFeatureEagerOnly(
        feature=feature_lib.Dataset({'int': dtype}),
        shape={'int': ()},  # shape of each element of the dataset
        dtype={'int': dtype},
        serialized_info={
            'int': feature_lib.TensorInfo(shape=(None,), dtype=dtype),
        },
        tests=[
            # Python array
            testing.FeatureExpectationItem(
                value=[{'int': 1}, {'int': 2}, {'int': 3}],
                expected=tf.data.Dataset.from_tensor_slices({'int': [1, 2, 3]}),
            ),
            # Numpy array
            testing.FeatureExpectationItem(
                value=dataset_utils.as_numpy(
                    tf.data.Dataset.from_tensor_slices(
                        {'int': np.ones(shape=(3,), dtype=np.int32)}
                    )
                ),
                expected=tf.data.Dataset.from_tensor_slices({'int': [1, 1, 1]}),
            ),
            # Dataset length doesn't matter
            testing.FeatureExpectationItem(
                value=dataset_utils.as_numpy(
                    tf.data.Dataset.from_tensor_slices(
                        {'int': np.ones(shape=(4,), dtype=np.int32)}
                    )
                ),
                expected=tf.data.Dataset.from_tensor_slices(
                    {'int': [1, 1, 1, 1]}
                ),
            ),
        ],
        test_attributes=dict(_length=None),
    )

  @parameterized.parameters((np.int64), (tf.int64))
  def test_label(self, dtype):
    self.assertFeatureEagerOnly(
        feature=feature_lib.Dataset(
            {
                'label': feature_lib.ClassLabel(names=['left', 'right']),
            },
            length=None,
        ),
        shape={'label': ()},
        dtype={'label': dtype},
        serialized_info={
            'label': feature_lib.TensorInfo(shape=(None,), dtype=dtype),
        },
        tests=[
            testing.FeatureExpectationItem(
                value=[
                    {'label': 'right'},
                    {'label': 'left'},
                    {'label': 'left'},
                ],
                expected=tf.data.Dataset.from_tensor_slices(
                    {'label': [1, 0, 0]}
                ),
            ),
            # Variable sequence length
            testing.FeatureExpectationItem(
                value=dataset_utils.as_numpy(
                    tf.data.Dataset.from_tensor_slices(
                        {'label': ['right', 'left', 'right', 'left']}
                    )
                ),
                expected=tf.data.Dataset.from_tensor_slices(
                    {'label': [1, 0, 1, 0]}
                ),
            ),
        ],
        test_attributes=dict(_length=None),
    )

  @parameterized.parameters(
      (np.object_, np.int32, np.uint8),
      (tf.string, tf.int32, tf.uint8),
  )
  def test_nested(self, a_dtype, bc_dtype, bd_dtype):
    self.assertFeatureEagerOnly(
        feature=feature_lib.Dataset(
            {
                'a': a_dtype,
                'b': {
                    'c': feature_lib.Tensor(shape=(4, 2), dtype=bc_dtype),
                    'd': bd_dtype,
                },
            },
            length=None,
        ),
        shape={
            'a': (),
            'b': {
                'c': (4, 2),
                'd': (),
            },
        },
        dtype={
            'a': np.object_,
            'b': {
                'c': np.int32,
                'd': np.uint8,
            },
        },
        tests=[
            testing.FeatureExpectationItem(
                value=dataset_utils.as_numpy(
                    tf.data.Dataset.from_tensor_slices({
                        'a': ['aa', 'b', 'ccc'],
                        'b': {
                            'c': np.ones(shape=(3, 4, 2), dtype=np.int32),
                            'd': [1, 2, 3],
                        },
                    })
                ),
                expected=tf.data.Dataset.from_tensor_slices({
                    'a': [tf.compat.as_bytes(t) for t in ('aa', 'b', 'ccc')],
                    'b': {
                        'c': np.ones(shape=(3, 4, 2), dtype=np.int32),
                        'd': [1, 2, 3],
                    },
                }),
            ),
            testing.FeatureExpectationItem(
                value=dataset_utils.as_numpy(
                    tf.data.Dataset.from_tensor_slices({
                        'a': [str(i) for i in range(100)],
                        'b': {  # pylint: disable=g-complex-comprehension
                            'c': [
                                np.ones(shape=(4, 2), dtype=np.int32)
                                for _ in range(100)
                            ],
                            'd': [5 for _ in range(100)],
                        },
                    })
                ),
                expected=tf.data.Dataset.from_tensor_slices({
                    'a': [tf.compat.as_bytes(str(i)) for i in range(100)],
                    'b': {
                        'c': np.ones(shape=(100, 4, 2), dtype=np.int32),
                        'd': [5] * 100,
                    },
                }),
            ),
        ],
    )

  @parameterized.parameters(
      (np.object_, np.int32, np.uint8),
      (tf.string, tf.int32, tf.uint8),
  )
  def test_input_dict(self, a_dtype, bc_dtype, bd_dtype):
    self.assertFeatureEagerOnly(
        feature=feature_lib.Dataset(
            {
                'a': a_dtype,
                'b': {
                    'c': feature_lib.Tensor(shape=(4, 2), dtype=bc_dtype),
                    'd': bd_dtype,
                },
            },
            length=None,
        ),
        shape={
            'a': (),
            'b': {
                'c': (4, 2),
                'd': (),
            },
        },
        dtype={
            'a': np.object_,
            'b': {
                'c': np.int32,
                'd': np.uint8,
            },
        },
        tests=[
            testing.FeatureExpectationItem(
                value={
                    'a': ['aa', 'b', 'ccc'],
                    'b': {
                        'c': np.ones(shape=(3, 4, 2), dtype=np.int32),
                        'd': [1, 2, 3],
                    },
                },
                expected=tf.data.Dataset.from_tensor_slices({
                    'a': [tf.compat.as_bytes(t) for t in ('aa', 'b', 'ccc')],
                    'b': {
                        'c': np.ones(shape=(3, 4, 2), dtype=np.int32),
                        'd': [1, 2, 3],
                    },
                }),
            ),
            testing.FeatureExpectationItem(
                value={
                    'a': [str(i) for i in range(100)],
                    'b': {  # pylint: disable=g-complex-comprehension
                        'c': [
                            np.ones(shape=(4, 2), dtype=np.int32)
                            for _ in range(100)
                        ],
                        'd': [5 for _ in range(100)],
                    },
                },
                expected=tf.data.Dataset.from_tensor_slices({
                    'a': [tf.compat.as_bytes(str(i)) for i in range(100)],
                    'b': {
                        'c': np.ones(shape=(100, 4, 2), dtype=np.int32),
                        'd': [5] * 100,
                    },
                }),
            ),
            # Wrong length in one of the lists.
            testing.FeatureExpectationItem(
                value={
                    'a': ['aa'],
                    'b': {
                        'c': np.ones(shape=(3, 4, 2), dtype=np.int32),
                        'd': [1, 2, 3],
                    },
                },
                raise_cls=ValueError,
                raise_msg=(
                    'The length of all elements of one sequence should be the'
                    ' same.'
                ),
            ),
        ],
    )

  @parameterized.parameters(
      (np.object_, np.uint8),
      (tf.string, tf.uint8),
  )
  def test_decoding(self, a_dtype, bc_dtype):
    self.assertFeatureEagerOnly(
        feature=feature_lib.Dataset(
            {
                'a': a_dtype,
                'b': {
                    'c': bc_dtype,
                },
            },
            length=None,
        ),
        shape={
            'a': (),
            'b': {
                'c': (),
            },
        },
        dtype={
            'a': np.object_,
            'b': {
                'c': np.uint8,
            },
        },
        tests=[
            testing.FeatureExpectationItem(
                value=dataset_utils.as_numpy(
                    tf.data.Dataset.from_tensor_slices({
                        'a': ['aa', 'b', 'ccc'],
                        'b': {
                            'c': [1, 2, 3],
                        },
                    })
                ),
                decoders={
                    'b': {
                        'c': IncrementDecoder(),
                    },
                },
                expected=tf.data.Dataset.from_tensor_slices({
                    'a': [tf.compat.as_bytes(t) for t in ('aa', 'b', 'ccc')],
                    'b': {
                        'c': [2, 3, 4],
                    },
                }),
            ),
        ],
    )


class DatasetFeatureTest(
    parameterized.TestCase, testing.FeatureExpectationsTestCase
):

  @parameterized.parameters((np.int32), (tf.int32))
  def test_int(self, dtype):
    self.assertFeatureEagerOnly(
        feature=feature_lib.Dataset(dtype, length=3),
        shape=(),
        dtype=np.int32,
        tests=[
            # Python array
            testing.FeatureExpectationItem(
                value=[1, 2, 3],
                expected=tf.data.Dataset.from_tensor_slices([1, 2, 3]),
            ),
            # Numpy array
            testing.FeatureExpectationItem(
                value=np.ones(shape=(3,), dtype=np.int32),
                expected=tf.data.Dataset.from_tensor_slices([1, 1, 1]),
            ),
            # Datasets with a different lenght will fail on encoding.
            testing.FeatureExpectationItem(
                value=np.ones(shape=(4,), dtype=np.int32),
                raise_cls=ValueError,
                raise_msg='Error while serializing feature',
            ),
        ],
    )

  def test_label(self):
    self.assertFeatureEagerOnly(
        feature=feature_lib.Dataset(
            feature_lib.ClassLabel(names=['left', 'right']),
        ),
        shape=(),
        dtype=np.int64,
        tests=[
            testing.FeatureExpectationItem(
                value=['right', 'left', 'left'],
                expected=tf.data.Dataset.from_tensor_slices([1, 0, 0]),
            ),
            # Variable sequence length
            testing.FeatureExpectationItem(
                value=['right', 'left', 'right', 'left'],
                expected=tf.data.Dataset.from_tensor_slices([1, 0, 1, 0]),
            ),
            # Empty sequence length
            testing.FeatureExpectationItem(
                value=[],
                expected=[],
            ),
        ],
    )

  def test_getattr(self):
    feature = feature_lib.Dataset(
        feature_lib.ClassLabel(names=['left', 'right']),
    )
    self.assertEqual(feature.names, ['left', 'right'])

    feature = feature_lib.Dataset(
        {
            'label': feature_lib.ClassLabel(names=['left', 'right']),
        }
    )
    self.assertEqual(feature['label'].names, ['left', 'right'])

  def test_metadata(self):
    feature = feature_lib.Dataset(feature_lib.ClassLabel(num_classes=2))
    feature.feature.names = ['left', 'right']
    with testing.tmp_dir() as tmp_dir:
      feature.save_metadata(data_dir=tmp_dir, feature_name='test')

      feature2 = feature_lib.Dataset(feature_lib.ClassLabel(num_classes=2))
      feature2.load_metadata(data_dir=tmp_dir, feature_name='test')
    self.assertEqual(feature2.feature.names, ['left', 'right'])


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the sequential writer."""
import tempfile

from absl.testing import absltest
from absl.testing import parameterized
import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import sequential_writer


def _generate_split(num_examples: int) -> tf.data.Dataset:
  data = {
      'a': list(range(num_examples)),
      'b': list(range(num_examples)),
  }
  return tf.data.Dataset.from_tensor_slices(data)


def _dataset_info(
    data_dir: str, name: str = 'test_dataset'
) -> dataset_info.DatasetInfo:
  return dataset_info.DatasetInfo(
      builder=dataset_info.DatasetIdentity(
          name=name,
          data_dir=data_dir,
          module_name='',
          version='1.0.0',
      ),
      description='Test builder',
      features=tfds.features.FeaturesDict({
          'a': tf.int32,
          'b': tf.int32,
      }),
  )


class SequentialWriterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      {
          'testcase_name': 'uneven_shards',
          'num_examples': 5,
          'max_examples_per_shard': 3,
          'num_expected_shards': 2,
      },
      {
          'testcase_name': 'even_shards',
          'num_examples': 6,
          'max_examples_per_shard': 3,
          'num_expected_shards': 2,
      },
      {
          'testcase_name': 'one_full_shard',
          'num_examples': 3,
          'max_examples_per_shard': 3,
          'num_expected_shards': 1,
      },
      {
          'testcase_name': 'one_not_full_shard',
          'num_examples': 2,
          'max_examples_per_shard': 3,
          'num_expected_shards': 1,
      },
  )
  def test_generates_correct_shards(
      self, num_examples, max_examples_per_shard, num_expected_shards
  ):
    data_dir = tempfile.mkdtemp('data_dir')
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir),
        max_examples_per_shard=max_examples_per_shard,
    )
    writer.initialize_splits(['train'])
    for example in tfds.as_numpy(_generate_split(num_examples)):
      writer.add_examples({'train': [example]})
    writer.close_splits(['train'])

    ds_builder = tfds.builder_from_directory(data_dir)
    self.assertEqual(
        ds_builder.info.splits['train'].num_shards, num_expected_shards
    )
    counter = (
        ds_builder.as_dataset(split='train')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, num_examples)

  def test_writes_multiple_examples(self):
    num_examples = 5
    max_examples_per_shard = 3
    num_expected_shards = 2
    data_dir = tempfile.mkdtemp('data_dir')

    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir),
        max_examples_per_shard=max_examples_per_shard,
    )
    writer.initialize_splits(['train'])
    writer.add_examples(
        {'train': list(tfds.as_numpy(_generate_split(num_examples)))}
    )
    writer.close_splits(['train'])

    ds_builder = tfds.builder_from_directory(data_dir)
    self.assertEqual(
        ds_builder.info.splits['train'].num_shards, num_expected_shards
    )
    counter = (
        ds_builder.as_dataset(split='train')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, num_examples)

  def test_writes_multiple_splits(self):
    num_examples_train = 5
    num_examples_eval = 7
    max_examples_per_shard = 3
    num_expected_shards_train = 2
    num_expected_shards_eval = 3

    data_dir = tempfile.mkdtemp('data_dir')

    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir),
        max_examples_per_shard=max_examples_per_shard,
    )
    writer.initialize_splits(['train', 'eval'])
    writer.add_examples({
        'train': list(tfds.as_numpy(_generate_split(num_examples_train))),
        'eval': list(tfds.as_numpy(_generate_split(num_examples_eval))),
    })
    writer.close_splits(['train', 'eval'])
    ds_builder = tfds.builder_from_directory(data_dir)
    self.assertEqual(
        ds_builder.info.splits['train'].num_shards, num_expected_shards_train
    )
    self.assertEqual(
        ds_builder.info.splits['eval'].num_shards, num_expected_shards_eval
    )
    counter = (
        ds_builder.as_dataset(split='train')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, num_examples_train)
    counter = (
        ds_builder.as_dataset(split='eval')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, num_examples_eval)

  def test_closes_all(self):
    num_examples_train = 5
    num_examples_eval = 7
    max_examples_per_shard = 3
    num_expected_shards_train = 2
    num_expected_shards_eval = 3

    data_dir = tempfile.mkdtemp('data_dir')

    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir),
        max_examples_per_shard=max_examples_per_shard,
    )
    writer.initialize_splits(['train', 'eval'])
    writer.add_examples({
        'train': list(tfds.as_numpy(_generate_split(num_examples_train))),
        'eval': list(tfds.as_numpy(_generate_split(num_examples_eval))),
    })
    writer.close_all()
    ds_builder = tfds.builder_from_directory(data_dir)
    self.assertEqual(
        ds_builder.info.splits['train'].num_shards, num_expected_shards_train
    )
    self.assertEqual(
        ds_builder.info.splits['eval'].num_shards, num_expected_shards_eval
    )
    counter = (
        ds_builder.as_dataset(split='train')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, num_examples_train)
    counter = (
        ds_builder.as_dataset(split='eval')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, num_examples_eval)

  def test_writes_multiple_splits_sequentially(self):
    num_examples_train = 5
    num_examples_eval = 7
    max_examples_per_shard = 3
    num_expected_shards_train = 2
    num_expected_shards_eval = 3

    data_dir = tempfile.mkdtemp('data_dir')

    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir),
        max_examples_per_shard=max_examples_per_shard,
    )
    writer.initialize_splits(['train'])
    writer.add_examples(
        {
            'train': list(tfds.as_numpy(_generate_split(num_examples_train))),
        }
    )
    writer.close_splits(['train'])
    writer.initialize_splits(['eval'])
    writer.add_examples(
        {
            'eval': list(tfds.as_numpy(_generate_split(num_examples_eval))),
        }
    )
    writer.close_splits(['eval'])
    ds_builder = tfds.builder_from_directory(data_dir)
    self.assertEqual(
        ds_builder.info.splits['train'].num_shards, num_expected_shards_train
    )
    self.assertEqual(
        ds_builder.info.splits['eval'].num_shards, num_expected_shards_eval
    )
    counter = (
        ds_builder.as_dataset(split='train')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, num_examples_train)
    counter = (
        ds_builder.as_dataset(split='eval')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, num_examples_eval)

  def test_fails_to_append_to_nonexisting_split(self):
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info('/unused/dir'), max_examples_per_shard=1
    )
    with self.assertRaises(KeyError):
      writer.add_examples(
          {
              'train': list(tfds.as_numpy(_generate_split(1))),
          }
      )

  def test_fails_to_close_a_nonexisting_split(self):
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info('/unused/dir'), max_examples_per_shard=1
    )
    with self.assertRaises(KeyError):
      writer.close_splits(['train'])

  def test_fails_to_initialize_same_split_twice(self):
    data_dir = tempfile.mkdtemp('data_dir')
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir), max_examples_per_shard=1
    )
    writer.initialize_splits(['train'])
    with self.assertRaises(KeyError):
      writer.initialize_splits(['train'])

  def test_initializes_same_split_twice(self):
    data_dir = tempfile.mkdtemp('data_dir')
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir), max_examples_per_shard=1
    )
    writer.initialize_splits(['train'])
    writer.initialize_splits(['train'], fail_if_exists=False)

  def test_fails_to_append_to_closed_split(self):
    data_dir = tempfile.mkdtemp('data_dir')
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir), max_examples_per_shard=1
    )
    writer.initialize_splits(['train'])
    writer.close_splits(['train'])
    with self.assertRaises(ValueError):
      writer.add_examples(
          {
              'train': list(tfds.as_numpy(_generate_split(1))),
          }
      )

  def test_closing_a_closed_split_is_a_noop(self):
    data_dir = tempfile.mkdtemp('data_dir')
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir), max_examples_per_shard=1
    )
    writer.initialize_splits(['train'])
    writer.close_splits(['train'])
    writer.close_splits(['train'])

  def test_append_to_different_dataset_fails(self):
    data_dir = tempfile.mkdtemp('data_dir')
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir), max_examples_per_shard=3
    )
    writer.initialize_splits(['train'])
    writer.close_splits(['train'])
    with self.assertRaises(ValueError):
      sequential_writer.SequentialWriter(
          _dataset_info(data_dir, 'new_name'),
          max_examples_per_shard=3,
          overwrite=False,
      )

  def test_append_to_non_existent_works(self):
    data_dir = tempfile.mkdtemp('data_dir')
    writer = sequential_writer.SequentialWriter(
        ds_info=_dataset_info(data_dir),
        max_examples_per_shard=3,
        overwrite=False,
    )
    writer.initialize_splits(['train'])
    writer.close_splits(['train'])

  def test_append_to_existing_dataset_works(self):
    data_dir = tempfile.mkdtemp('data_dir')
    ds_info = _dataset_info(data_dir)
    writer = sequential_writer.SequentialWriter(
        ds_info=ds_info, max_examples_per_shard=3
    )
    writer.initialize_splits(['train'])
    writer.add_examples({'train': list(tfds.as_numpy(_generate_split(4)))})
    writer.close_splits(
        ['train']
    )  # The split should have 2 shards with 4 examples

    new_writer = sequential_writer.SequentialWriter(
        ds_info=ds_info, max_examples_per_shard=2, overwrite=False
    )
    new_writer.initialize_splits(['train'], fail_if_exists=False)
    new_writer.add_examples({'train': list(tfds.as_numpy(_generate_split(3)))})
    new_writer.close_splits(['train'])  # We added 2 shards with 3 examples

    ds_builder = tfds.builder_from_directory(data_dir)
    self.assertEqual(ds_builder.info.splits['train'].num_shards, 4)
    counter = (
        ds_builder.as_dataset(split='train')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, 7)

  def test_overwrites(self):
    data_dir = tempfile.mkdtemp('data_dir')
    ds_info = _dataset_info(data_dir)
    writer = sequential_writer.SequentialWriter(
        ds_info=ds_info, max_examples_per_shard=3
    )
    writer.initialize_splits(['train'])
    writer.add_examples({'train': list(tfds.as_numpy(_generate_split(4)))})
    writer.close_splits(
        ['train']
    )  # The split should have 2 shards with 4 examples

    new_writer = sequential_writer.SequentialWriter(
        ds_info=ds_info, max_examples_per_shard=2, overwrite=True
    )
    new_writer.initialize_splits(['train'], fail_if_exists=True)
    new_writer.add_examples({'train': list(tfds.as_numpy(_generate_split(3)))})
    new_writer.close_splits(['train'])  # We added 2 shards with 3 examples

    ds_builder = tfds.builder_from_directory(data_dir)
    self.assertEqual(ds_builder.info.splits['train'].num_shards, 2)
    counter = (
        ds_builder.as_dataset(split='train')
        .reduce(0, lambda v, _: v + 1)
        .numpy()
    )
    self.assertEqual(counter, 3)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for movielens_parsing."""

import os

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.structured import movielens_parsing


class MovieLensUtilsTest(tfds.testing.TestCase):
  """Tests for helper functions in movielens_parsing."""

  def test_parse_current_movies_data(self):
    """Test for parse_current_movies_data."""
    latest_small_path = os.path.join(
        tfds.testing.fake_examples_dir(),
        'movielens',
        'ml-latest-small',
    )
    movies_generator = movielens_parsing.parse_current_movies_data(
        latest_small_path
    )
    expected_result = [
        (
            0,
            {
                'movie_id': '1',
                'movie_genres': ['Fantasy', 'War'],
                'movie_title': 'Horse Movie (2011)',
            },
        ),
        (
            1,
            {
                'movie_id': '2',
                'movie_genres': ['Adventure', 'Documentary'],
                'movie_title': 'Eagle Movie (1999)',
            },
        ),
        (
            2,
            {
                'movie_id': '3',
                'movie_genres': ['(no genres listed)'],
                'movie_title': 'Test Movie (2014)',
            },
        ),
        (
            3,
            {
                'movie_id': '4',
                'movie_genres': ['Fantasy', 'Horror', 'Musical'],
                'movie_title': 'Fake Movie (2007)',
            },
        ),
        (
            4,
            {
                'movie_id': '5',
                'movie_genres': ['Mystery'],
                'movie_title': 'Squirrel Movie (2017)',
            },
        ),
        (
            5,
            {
                'movie_id': '6',
                'movie_genres': ['Drama', 'War'],
                'movie_title': 'Test Movie (2003)',
            },
        ),
        (
            6,
            {
                'movie_id': '7',
                'movie_genres': ['Horror'],
                'movie_title': 'Rabbit Movie (1999)',
            },
        ),
        (
            7,
            {
                'movie_id': '8',
                'movie_genres': ['Mystery', 'War'],
                'movie_title': 'Tree Movie (2001)',
            },
        ),
    ]
    parsed_example_list = list(movies_generator)
    self.assertEqual(expected_result, parsed_example_list)

  def test_parse_current_ratings_data(self):
    """Test for parse_current_ratings_data."""
    fake_dir_path = os.path.join(
        tfds.testing.fake_examples_dir(),
        'movielens',
        'ml-latest-small',
    )
    movies_generator = movielens_parsing.parse_current_ratings_data(
        fake_dir_path
    )
    expected_result = [
        (
            0,
            {
                'movie_id': '5',
                'movie_genres': ['Mystery'],
                'movie_title': 'Squirrel Movie (2017)',
                'user_id': '1',
                'user_rating': '2.5',
                'timestamp': '1329800589',
            },
        ),
        (
            1,
            {
                'movie_id': '2',
                'movie_genres': ['Adventure', 'Documentary'],
                'movie_title': 'Eagle Movie (1999)',
                'user_id': '2',
                'user_rating': '1.5',
                'timestamp': '1128006292',
            },
        ),
        (
            2,
            {
                'movie_id': '3',
                'movie_genres': ['(no genres listed)'],
                'movie_title': 'Test Movie (2014)',
                'user_id': '2',
                'user_rating': '3.0',
                'timestamp': '1029868117',
            },
        ),
        (
            3,
            {
                'movie_id': '4',
                'movie_genres': ['Fantasy', 'Horror', 'Musical'],
                'movie_title': 'Fake Movie (2007)',
                'user_id': '2',
                'user_rating': '0.5',
                'timestamp': '1589647790',
            },
        ),
        (
            4,
            {
                'movie_id': '5',
                'movie_genres': ['Mystery'],
                'movie_title': 'Squirrel Movie (2017)',
                'user_id': '2',
                'user_rating': '5.0',
                'timestamp': '1553439546',
            },
        ),
        (
            5,
            {
                'movie_id': '1',
                'movie_genres': ['Fantasy', 'War'],
                'movie_title': 'Horse Movie (2011)',
                'user_id': '3',
                'user_rating': '1.5',
                'timestamp': '1580974707',
            },
        ),
        (
            6,
            {
                'movie_id': '2',
                'movie_genres': ['Adventure', 'Documentary'],
                'movie_title': 'Eagle Movie (1999)',
                'user_id': '3',
                'user_rating': '2.0',
                'timestamp': '820693640',
            },
        ),
        (
            7,
            {
                'movie_id': '4',
                'movie_genres': ['Fantasy', 'Horror', 'Musical'],
                'movie_title': 'Fake Movie (2007)',
                'user_id': '4',
                'user_rating': '1.0',
                'timestamp': '1496507328',
            },
        ),
    ]
    parsed_example_list = list(movies_generator)
    self.assertEqual(expected_result, parsed_example_list)

  def test_parse_100k_movies_data(self):
    """Test for parse_100k_movies_data."""
    fake_dir_path = os.path.join(
        tfds.testing.fake_examples_dir(),
        'movielens',
        'ml-100k',
    )
    movies_generator = movielens_parsing.parse_100k_movies_data(fake_dir_path)
    expected_result = [
        (
            0,
            {
                'movie_id': '1',
                'movie_genres': ['Unknown', 'Children'],
                'movie_title': 'Fake Movie (1996)',
            },
        ),
        (
            1,
            {
                'movie_id': '2',
                'movie_genres': ['Western'],
                'movie_title': 'Dog Movie (1994)',
            },
        ),
        (
            2,
            {
                'movie_id': '3',
                'movie_genres': ['Adventure'],
                'movie_title': 'Non-existent Movie (1995)',
            },
        ),
        (
            3,
            {
                'movie_id': '4',
                'movie_genres': ['Documentary', 'Film-Noir', 'Musical'],
                'movie_title': 'Owl Movie (1997)',
            },
        ),
        (
            4,
            {
                'movie_id': '5',
                'movie_genres': ['Crime', 'Musical'],
                'movie_title': 'Cat Movie (1998)',
            },
        ),
        (
            5,
            {
                'movie_id': '6',
                'movie_genres': ['Mystery', 'Romance'],
                'movie_title': 'Test Movie (2003)',
            },
        ),
        (
            6,
            {
                'movie_id': '7',
                'movie_genres': ['Action', 'Comedy', 'Fantasy', 'Romance'],
                'movie_title': 'Rabbit Movie (1999)',
            },
        ),
        (
            7,
            {
                'movie_id': '8',
                'movie_genres': ['Documentary'],
                'movie_title': 'Tree Movie (2001)',
            },
        ),
    ]
    parsed_example_list = list(movies_generator)
    self.assertEqual(expected_result, parsed_example_list)

  def test_parse_100k_ratings_data(self):
    """Test for parse_100k_ratings_data."""
    fake_dir_path = os.path.join(
        tfds.testing.fake_examples_dir(),
        'movielens',
        'ml-100k',
    )
    movies_generator = movielens_parsing.parse_100k_ratings_data(fake_dir_path)
    expected_result = [
        (
            0,
            {
                'movie_id': '1',
                'movie_title': 'Fake Movie (1996)',
                'movie_genres': ['Unknown', 'Children'],
                'user_id': '1',
                'user_rating': '1',
                'timestamp': '867969113',
                'user_gender': False,
                'bucketized_user_age': 50,
                'user_occupation_text': 'technician',
                'user_occupation_label': 'technician/engineer',
                'user_zip_code': '10278',
                'raw_user_age': 54,
            },
        ),
        (
            1,
            {
                'movie_id': '2',
                'movie_genres': ['Western'],
                'movie_title': 'Dog Movie (1994)',
                'user_id': '1',
                'user_rating': '1',
                'timestamp': '1058848127',
                'user_gender': False,
                'bucketized_user_age': 50,
                'user_occupation_text': 'technician',
                'user_occupation_label': 'technician/engineer',
                'user_zip_code': '10278',
                'raw_user_age': 54,
            },
        ),
        (
            2,
            {
                'movie_id': '4',
                'movie_genres': ['Documentary', 'Film-Noir', 'Musical'],
                'movie_title': 'Owl Movie (1997)',
                'user_id': '1',
                'user_rating': '3',
                'timestamp': '997136856',
                'user_gender': False,
                'bucketized_user_age': 50,
                'user_occupation_text': 'technician',
                'user_occupation_label': 'technician/engineer',
                'user_zip_code': '10278',
                'raw_user_age': 54,
            },
        ),
        (
            3,
            {
                'movie_id': '2',
                'movie_genres': ['Western'],
                'movie_title': 'Dog Movie (1994)',
                'user_id': '2',
                'user_rating': '4',
                'timestamp': '1188937963',
                'user_gender': True,
                'bucketized_user_age': 1,
                'user_occupation_text': 'other',
                'user_occupation_label': 'other/not specified',
                'user_zip_code': '34729',
                'raw_user_age': 16,
            },
        ),
        (
            4,
            {
                'movie_id': '1',
                'movie_genres': ['Unknown', 'Children'],
                'movie_title': 'Fake Movie (1996)',
                'user_id': '3',
                'user_rating': '5',
                'timestamp': '1083844839',
                'user_gender': True,
                'bucketized_user_age': 25,
                'user_occupation_text': 'librarian',
                'user_occupation_label': 'librarian',
                'user_zip_code': '20112',
                'raw_user_age': 29,
            },
        ),
        (
            5,
            {
                'movie_id': '5',
                'movie_genres': ['Crime', 'Musical'],
                'movie_title': 'Cat Movie (1998)',
                'user_id': '3',
                'user_rating': '3',
                'timestamp': '1042582463',
                'user_gender': True,
                'bucketized_user_age': 25,
                'user_occupation_text': 'librarian',
                'user_occupation_label': 'librarian',
                'user_zip_code': '20112',
                'raw_user_age': 29,
            },
        ),
        (
            6,
            {
                'movie_id': '3',
                'movie_genres': ['Adventure'],
                'movie_title': 'Non-existent Movie (1995)',
                'user_id': '4',
                'user_rating': '2',
                'timestamp': '803797344',
                'user_gender': False,
                'bucketized_user_age': 45,
                'user_occupation_text': 'none',
                'user_occupation_label': 'other/not specified',
                'user_zip_code': '09583',
                'raw_user_age': 47,
            },
        ),
        (
            7,
            {
                'movie_id': '4',
                'movie_genres': ['Documentary', 'Film-Noir', 'Musical'],
                'movie_title': 'Owl Movie (1997)',
                'user_id': '4',
                'user_rating': '1',
                'timestamp': '1235956810',
                'user_gender': False,
                'bucketized_user_age': 45,
                'user_occupation_text': 'none',
                'user_occupation_label': 'other/not specified',
                'user_zip_code': '09583',
                'raw_user_age': 47,
            },
        ),
    ]
    parsed_example_list = list(movies_generator)
    self.assertEqual(expected_result, parsed_example_list)

  def test_parse_1m_movies_data(self):
    """Test for parse_1m_movies_data."""
    fake_dir_path = os.path.join(
        tfds.testing.fake_examples_dir(),
        'movielens',
        'ml-1m',
    )
    movies_generator = movielens_parsing.parse_1m_movies_data(fake_dir_path)
    expected_result = [
        (
            0,
            {
                'movie_id': '1',
                'movie_genres': ['Fantasy', 'Children'],
                'movie_title': 'Fake Movie (2005)',
            },
        ),
        (
            1,
            {
                'movie_id': '2',
                'movie_genres': ['Crime', 'Sci-Fi', 'Western'],
                'movie_title': 'Dog Movie (1994)',
            },
        ),
        (
            2,
            {
                'movie_id': '3',
                'movie_genres': ['Film-Noir'],
                'movie_title': 'Non-existent Movie (2003)',
            },
        ),
        (
            3,
            {
                'movie_id': '4',
                'movie_genres': ['Comedy', 'Drama', 'Romance'],
                'movie_title': 'Owl Movie (2018)',
            },
        ),
        (
            4,
            {
                'movie_id': '5',
                'movie_genres': ['Comedy', 'Romance'],
                'movie_title': 'Cat Movie (2018)',
            },
        ),
        (
            5,
            {
                'movie_id': '6',
                'movie_genres': ['Drama', 'War'],
                'movie_title': 'Test Movie (2003)',
            },
        ),
        (
            6,
            {
                'movie_id': '7',
                'movie_genres': ['Horror'],
                'movie_title': 'Rabbit Movie (1999)',
            },
        ),
        (
            7,
            {
                'movie_id': '8',
                'movie_genres': ['Mystery', 'War'],
                'movie_title': 'Tree Movie (2001)',
            },
        ),
    ]
    parsed_example_list = list(movies_generator)
    self.assertEqual(expected_result, parsed_example_list)

  def test_parse_1m_ratings_data(self):
    """Test for parse_1m_ratings_data."""
    fake_dir_path = os.path.join(
        tfds.testing.fake_examples_dir(),
        'movielens',
        'ml-1m',
    )
    movies_generator = movielens_parsing.parse_1m_ratings_data(fake_dir_path)
    expected_result = [
        (
            0,
            {
                'movie_id': '1',
                'movie_genres': ['Fantasy', 'Children'],
                'movie_title': 'Fake Movie (2005)',
                'user_id': '1',
                'user_rating': '1',
                'timestamp': '867969113',
                'user_gender': False,
                'bucketized_user_age': '50',
                'user_occupation_text': 'retired',
                'user_occupation_label': 'retired',
                'user_zip_code': '10278',
            },
        ),
        (
            1,
            {
                'movie_id': '2',
                'movie_genres': ['Crime', 'Sci-Fi', 'Western'],
                'movie_title': 'Dog Movie (1994)',
                'user_id': '1',
                'user_rating': '1',
                'timestamp': '1058848127',
                'user_gender': False,
                'bucketized_user_age': '50',
                'user_occupation_text': 'retired',
                'user_occupation_label': 'retired',
                'user_zip_code': '10278',
            },
        ),
        (
            2,
            {
                'movie_id': '4',
                'movie_genres': ['Comedy', 'Drama', 'Romance'],
                'movie_title': 'Owl Movie (2018)',
                'user_id': '1',
                'user_rating': '3',
                'timestamp': '997136856',
                'user_gender': False,
                'bucketized_user_age': '50',
                'user_occupation_text': 'retired',
                'user_occupation_label': 'retired',
                'user_zip_code': '10278',
            },
        ),
        (
            3,
            {
                'movie_id': '2',
                'movie_genres': ['Crime', 'Sci-Fi', 'Western'],
                'movie_title': 'Dog Movie (1994)',
                'user_id': '2',
                'user_rating': '4',
                'timestamp': '1188937963',
                'user_gender': True,
                'bucketized_user_age': '18',
                'user_occupation_text': 'college/grad student',
                'user_occupation_label': 'student',
                'user_zip_code': '34729',
            },
        ),
        (
            4,
            {
                'movie_id': '1',
                'movie_genres': ['Fantasy', 'Children'],
                'movie_title': 'Fake Movie (2005)',
                'user_id': '3',
                'user_rating': '5',
                'timestamp': '1083844839',
                'user_gender': True,
                'bucketized_user_age': '1',
                'user_occupation_text': 'scientist',
                'user_occupation_label': 'scientist',
                'user_zip_code': '20112',
            },
        ),
        (
            5,
            {
                'movie_id': '5',
                'movie_genres': ['Comedy', 'Romance'],
                'movie_title': 'Cat Movie (2018)',
                'user_id': '3',
                'user_rating': '3',
                'timestamp': '1042582463',
                'user_gender': True,
                'bucketized_user_age': '1',
                'user_occupation_text': 'scientist',
                'user_occupation_label': 'scientist',
                'user_zip_code': '20112',
            },
        ),
        (
            6,
            {
                'movie_id': '3',
                'movie_genres': ['Film-Noir'],
                'movie_title': 'Non-existent Movie (2003)',
                'user_id': '4',
                'user_rating': '2',
                'timestamp': '803797344',
                'user_gender': False,
                'bucketized_user_age': '56',
                'user_occupation_text': 'other/not specified',
                'user_occupation_label': 'other/not specified',
                'user_zip_code': '09583',
            },
        ),
        (
            7,
            {
                'movie_id': '4',
                'movie_genres': ['Comedy', 'Drama', 'Romance'],
                'movie_title': 'Owl Movie (2018)',
                'user_id': '4',
                'user_rating': '1',
                'timestamp': '1235956810',
                'user_gender': False,
                'bucketized_user_age': '56',
                'user_occupation_text': 'other/not specified',
                'user_occupation_label': 'other/not specified',
                'user_zip_code': '09583',
            },
        ),
    ]
    parsed_example_list = list(movies_generator)
    self.assertEqual(expected_result, parsed_example_list)


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
"""Tests for tensorflow_datasets.core.features.tensor_feature."""

from absl.testing import parameterized
import jax.numpy as jnp
import numpy as np
import pytest
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import features as features_lib


class FeatureTensorTest(
    testing.FeatureExpectationsTestCase,
    parameterized.TestCase,
):

  @parameterized.parameters([
      (np.float32, features_lib.Encoding.NONE),
      (tf.float32, features_lib.Encoding.NONE),
      (np.float32, features_lib.Encoding.ZLIB),
      (tf.float32, features_lib.Encoding.ZLIB),
      (np.float32, features_lib.Encoding.BYTES),
      (tf.float32, features_lib.Encoding.BYTES),
  ])
  def test_shape_static(self, dtype, encoding: features_lib.Encoding):
    np_input = np.random.rand(2, 3).astype(np.float32)
    array_input = [
        [1, 2, 3],
        [4, 5, 6],
    ]

    self.assertFeature(
        feature=features_lib.Tensor(
            shape=(2, 3),
            dtype=dtype,
            encoding=encoding,
        ),
        dtype=dtype,
        shape=(2, 3),
        tests=[
            # Np array
            testing.FeatureExpectationItem(
                value=np_input,
                expected=np_input,
                expected_np=np_input,
            ),
            # Python array
            testing.FeatureExpectationItem(
                value=array_input,
                expected=array_input,
                expected_np=array_input,
            ),
            # Invalid dtype
            testing.FeatureExpectationItem(
                # On Windows, np default dtype is `int32`
                value=np.random.randint(256, size=(2, 3), dtype=np.int64),
                raise_cls=ValueError,
                raise_cls_np=ValueError,
                raise_msg='int64 do not match',
            ),
            # Invalid shape
            testing.FeatureExpectationItem(
                value=np.random.rand(2, 4).astype(np.float32),
                raise_cls=ValueError,
                raise_cls_np=ValueError,
                raise_msg='are incompatible',
            ),
        ],
        test_attributes={
            '_encoding': encoding,
        },
    )

  @parameterized.parameters([
      (np.int32, features_lib.Encoding.NONE),
      (tf.int32, features_lib.Encoding.NONE),
      (np.int32, features_lib.Encoding.ZLIB),
      (tf.int32, features_lib.Encoding.ZLIB),
      (np.int32, features_lib.Encoding.BYTES),
      (tf.int32, features_lib.Encoding.BYTES),
  ])
  def test_shape_dynamic(self, dtype, encoding: features_lib.Encoding):
    np_input_dynamic_1 = np.random.randint(256, size=(2, 3, 2), dtype=np.int32)
    np_input_dynamic_2 = np.random.randint(256, size=(5, 3, 2), dtype=np.int32)

    self.assertFeature(
        feature=features_lib.Tensor(
            shape=(None, 3, 2),
            dtype=dtype,
            encoding=encoding,
        ),
        dtype=dtype,
        shape=(None, 3, 2),
        tests=[
            testing.FeatureExpectationItem(
                value=np_input_dynamic_1,
                expected=np_input_dynamic_1,
            ),
            testing.FeatureExpectationItem(
                value=np_input_dynamic_2,
                expected=np_input_dynamic_2,
            ),
            # Invalid shape
            testing.FeatureExpectationItem(
                value=np.random.randint(256, size=(2, 3, 1), dtype=np.int32),
                raise_cls=ValueError,
                raise_cls_np=ValueError,
                raise_msg='are incompatible',
            ),
        ],
    )

  @parameterized.parameters([
      (np.int32, features_lib.Encoding.NONE),
      (tf.int32, features_lib.Encoding.NONE),
      (np.int32, features_lib.Encoding.ZLIB),
      (tf.int32, features_lib.Encoding.ZLIB),
      (np.int32, features_lib.Encoding.BYTES),
      (tf.int32, features_lib.Encoding.BYTES),
  ])
  def test_shape_dynamic_none_second(
      self, dtype, encoding: features_lib.Encoding
  ):
    np_input_dynamic_1 = np.random.randint(256, size=(3, 2, 2), dtype=np.int32)
    np_input_dynamic_2 = np.random.randint(256, size=(3, 5, 2), dtype=np.int32)

    self.assertFeature(
        feature=features_lib.Tensor(
            shape=(3, None, 2),  # None not at the first position.
            dtype=dtype,
            encoding=encoding,
        ),
        dtype=dtype,
        shape=(3, None, 2),
        tests=[
            testing.FeatureExpectationItem(
                value=np_input_dynamic_1,
                expected=np_input_dynamic_1,
            ),
            testing.FeatureExpectationItem(
                value=np_input_dynamic_2,
                expected=np_input_dynamic_2,
            ),
            # Invalid shape
            testing.FeatureExpectationItem(
                value=np.random.randint(256, size=(2, 3, 1), dtype=np.int32),
                raise_cls=ValueError,
                raise_cls_np=ValueError,
                raise_msg='are incompatible',
            ),
        ],
    )

  @parameterized.parameters([
      # (features_lib.Encoding.NONE,),  # Multiple unknown dims requires bytes
      (np.uint8, features_lib.Encoding.ZLIB),
      (tf.uint8, features_lib.Encoding.ZLIB),
      (np.uint8, features_lib.Encoding.BYTES),
      (tf.uint8, features_lib.Encoding.BYTES),
  ])
  def test_features_shape_dynamic_multi_none(
      self, dtype, encoding: features_lib.Encoding
  ):
    x = np.random.randint(256, size=(2, 3, 1), dtype=np.uint8)
    x_other_shape = np.random.randint(256, size=(4, 4, 1), dtype=np.uint8)
    wrong_shape = np.random.randint(256, size=(2, 3, 2), dtype=np.uint8)

    self.assertFeature(
        feature=features_lib.Tensor(
            shape=(None, None, 1),
            dtype=dtype,
            encoding=encoding,
        ),
        shape=(None, None, 1),
        dtype=dtype,
        tests=[
            testing.FeatureExpectationItem(
                value=x,
                expected=x,
            ),
            testing.FeatureExpectationItem(
                value=x_other_shape,
                expected=x_other_shape,
            ),
            testing.FeatureExpectationItem(
                value=wrong_shape,  # Wrong shape
                raise_cls=ValueError,
                raise_cls_np=ValueError,
                raise_msg='are incompatible',
            ),
        ],
    )

  @parameterized.parameters([
      # NONE only support single unknown dim, not 2.
      (np.uint8, features_lib.Encoding.BYTES, (2, None, 1)),
      (tf.uint8, features_lib.Encoding.BYTES, (2, None, 1)),
      (np.uint8, features_lib.Encoding.ZLIB, (None, None, 1)),
      (tf.uint8, features_lib.Encoding.ZLIB, (None, None, 1)),
      (np.uint8, features_lib.Encoding.BYTES, (None, None, 1)),
      (tf.uint8, features_lib.Encoding.BYTES, (None, None, 1)),
  ])
  def test_features_multi_none_sequence(
      self,
      dtype,
      encoding: features_lib.Encoding,
      shape,
  ):
    x = np.random.randint(256, size=(3, 2, 3, 1), dtype=np.uint8)
    x_other_shape = np.random.randint(256, size=(3, 2, 2, 1), dtype=np.uint8)

    self.assertFeature(
        feature=features_lib.Sequence(
            features_lib.Tensor(
                shape=shape,
                dtype=dtype,
                encoding=encoding,
            ),
        ),
        shape=(None,) + shape,
        dtype=dtype,
        tests=[
            testing.FeatureExpectationItem(
                value=x,
                expected=x,
                expected_np=x,
            ),
            testing.FeatureExpectationItem(
                value=x_other_shape,
                expected=x_other_shape,
                expected_np=x_other_shape,
            ),
            # TODO(epot): Is there a way to catch if the user try to encode
            # tensors with different shapes ?
        ],
    )

  @parameterized.parameters([
      (np.bool_, features_lib.Encoding.NONE),
      (tf.bool, features_lib.Encoding.NONE),
      (np.bool_, features_lib.Encoding.ZLIB),
      (tf.bool, features_lib.Encoding.ZLIB),
      (np.bool_, features_lib.Encoding.BYTES),
      (tf.bool, features_lib.Encoding.BYTES),
  ])
  def test_bool_flat(self, dtype, encoding: features_lib.Encoding):
    self.assertFeature(
        feature=features_lib.Tensor(
            shape=(),
            dtype=dtype,
            encoding=encoding,
        ),
        dtype=dtype,
        shape=(),
        tests=[
            testing.FeatureExpectationItem(
                value=np.array(True),
                expected=True,
                expected_np=True,
            ),
            testing.FeatureExpectationItem(
                value=np.array(False),
                expected=False,
                expected_np=False,
            ),
            testing.FeatureExpectationItem(
                value=True,
                expected=True,
                expected_np=True,
            ),
            testing.FeatureExpectationItem(
                value=False,
                expected=False,
                expected_np=False,
            ),
            testing.FeatureExpectationItem(
                value=1,
                expected=True,
                expected_np=True,
            ),
            testing.FeatureExpectationItem(
                value=2,
                expected=True,
                expected_np=True,
            ),
            testing.FeatureExpectationItem(
                value=0,
                expected=False,
                expected_np=False,
            ),
        ],
    )

  @parameterized.parameters([
      (np.bool_, features_lib.Encoding.NONE),
      (tf.bool, features_lib.Encoding.NONE),
      (np.bool_, features_lib.Encoding.ZLIB),
      (tf.bool, features_lib.Encoding.ZLIB),
      (np.bool_, features_lib.Encoding.BYTES),
      (tf.bool, features_lib.Encoding.BYTES),
  ])
  def test_bool_array(self, dtype, encoding: features_lib.Encoding):
    self.assertFeature(
        feature=features_lib.Tensor(
            shape=(3,),
            dtype=dtype,
            encoding=encoding,
        ),
        dtype=np.bool_,
        shape=(3,),
        tests=[
            testing.FeatureExpectationItem(
                value=np.array([True, True, False]),
                expected=[True, True, False],
                expected_np=[True, True, False],
            ),
            testing.FeatureExpectationItem(
                value=[True, False, True],
                expected=[True, False, True],
                expected_np=[True, False, True],
            ),
        ],
    )

  @parameterized.parameters(
      [
          (np.str_, features_lib.Encoding.NONE),
          (np.object_, features_lib.Encoding.NONE),
          (tf.string, features_lib.Encoding.NONE),
          # Bytes encoding not supported for tf.string
      ]
  )
  def test_string(self, dtype, encoding: features_lib.Encoding):
    nonunicode_text = 'hello world'
    unicode_text = '你好'

    self.assertFeature(
        feature=features_lib.Tensor(
            shape=(),
            dtype=dtype,
            encoding=encoding,
        ),
        shape=(),
        dtype=np.object_,
        tests=[
            # Non-unicode
            testing.FeatureExpectationItem(
                value=nonunicode_text,
                expected=tf.compat.as_bytes(nonunicode_text),
                expected_np=tf.compat.as_bytes(nonunicode_text),
            ),
            # Unicode
            testing.FeatureExpectationItem(
                value=unicode_text,
                expected=tf.compat.as_bytes(unicode_text),
                expected_np=tf.compat.as_bytes(unicode_text),
            ),
            # Empty string
            testing.FeatureExpectationItem(
                value='',
                expected=b'',
                expected_np=b'',
            ),
            # Trailing zeros
            testing.FeatureExpectationItem(
                value=b'abc\x00\x00',
                expected=b'abc\x00\x00',
                expected_np=b'abc\x00\x00',
            ),
        ],
    )

    self.assertFeature(
        feature=features_lib.Tensor(
            shape=(2, 1),
            dtype=dtype,
            encoding=encoding,
        ),
        shape=(2, 1),
        dtype=np.object_,
        tests=[
            testing.FeatureExpectationItem(
                value=[[nonunicode_text], [unicode_text]],
                expected=[
                    [tf.compat.as_bytes(nonunicode_text)],
                    [tf.compat.as_bytes(unicode_text)],
                ],
                expected_np=[
                    [tf.compat.as_bytes(nonunicode_text)],
                    [tf.compat.as_bytes(unicode_text)],
                ],
            ),
            testing.FeatureExpectationItem(
                value=[nonunicode_text, unicode_text],  # Wrong shape
                raise_cls=ValueError,
                raise_cls_np=ValueError,
                raise_msg='(2,) and (2, 1) must have the same rank',
            ),
            testing.FeatureExpectationItem(
                value=[['some text'], [123]],  # Wrong dtype
                raise_cls=TypeError,
                raise_cls_np=TypeError,
                raise_msg='Expected binary or unicode string, got 123',
            ),
        ],
    )


def test_invalid_input():
  with pytest.raises(ValueError, match='Multiple unknown dimensions'):
    features_lib.Tensor(
        shape=(2, None, None),
        dtype=np.uint8,
    ).encode_example(None)

  with pytest.raises(ValueError, match='Multiple unknown dimensions'):
    features_lib.Tensor(
        shape=(2, None, None),
        dtype=tf.uint8,
    ).encode_example(None)

  with pytest.raises(
      NotImplementedError,
      match='does not support `encoding=` when dtype is string',
  ):
    features_lib.Tensor(
        shape=(2, 1),
        dtype=np.str_,
        encoding=features_lib.Encoding.BYTES,
    )

  with pytest.raises(
      NotImplementedError,
      match='does not support `encoding=` when dtype is string',
  ):
    features_lib.Tensor(
        shape=(2, 1),
        dtype=tf.string,
        encoding=features_lib.Encoding.BYTES,
    )

  with pytest.raises(
      TypeError, match='converting it to `bool` will always output `True`'
  ):
    features_lib.Tensor(
        shape=(),
        dtype=np.bool_,
        encoding=features_lib.Encoding.BYTES,
    ).encode_example('0')

  with pytest.raises(
      TypeError, match='converting it to `bool` will always output `True`'
  ):
    features_lib.Tensor(
        shape=(),
        dtype=np.bool_,
        encoding=features_lib.Encoding.BYTES,
    ).encode_example('False')


def test_jax_bfloat16():
  features = features_lib.FeaturesDict(
      {
          'data': features_lib.Tensor(
              shape=(1,), dtype=tf.bfloat16, encoding='bytes'
          )
      }
  )
  data = {
      'data': jnp.array([6.0], dtype=jnp.bfloat16),
  }
  features.encode_example(data)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base register class."""

import abc
from typing import Any, Iterable, List, Type

from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import naming


class BaseRegister(abc.ABC):
  """Abstract register class.

  The register class indicates how to load a dataset from the name. Abstract
  methods `register.xyz()` have the same API as `tfds.xyz()`.

  ```
  register = MyRegister()
  register.list_builders()
  builder = register.builder('kaggle', 'my_dataset')
  ```

  Subclasses indicates how dataset code are fetched:

  * DataDirRegister: Find the dataset by looking at pre-generated dataset
    inside `data_dir`.
  * RemoteRegister: Find the dataset by fetching remote generation script.
  """

  @abc.abstractmethod
  def list_builders(self) -> List[str]:
    """Returns the list of registered builders.

    Returns:
      builder_names: The sorted, canonical list of builder names (including
        the eventual namespace). Example: `['kaggle:ds0', 'kaggle:ds1',...]`
    """
    raise NotImplementedError

  @abc.abstractmethod
  def list_dataset_references(self) -> Iterable[naming.DatasetReference]:
    """Yields registered datasets.

    Yields:
      references to datasets in this register. Note that configs and versions of
      the dataset are not included.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def builder_cls(
      self,
      name: naming.DatasetName,
  ) -> Type[dataset_builder.DatasetBuilder]:
    """Returns the `tfds.core.DatasetBuilder` instance.

    Contrary to `tfds.builder_cls`, `builder_name` here is already normalized (
    `my_dataset/my_config` -> `my_dataset`, `{'config': 'my_config'}`).

    Args:
      name: Builder name (e.g. `DatasetName('kaggle:mnist')`)

    Returns:
      builder_cls
    """
    raise NotImplementedError

  @abc.abstractmethod
  def builder(
      self,
      name: naming.DatasetName,
      **builder_kwargs: Any,
  ) -> dataset_builder.DatasetBuilder:
    """Returns the `tfds.core.DatasetBuilder` instance.

    Contrary to `tfds.builder`, `builder_name` here is already normalized (
    `my_dataset/my_config` -> `my_dataset`, `{'config': 'my_config'}`).

    Args:
      name: Builder name (e.g. `DatasetName('kaggle:mnist')`)
      **builder_kwargs: Additional kwargs forwarded to
        `tfds.core.DatasetBuilder` (version, config,...)

    Returns:
      builder
    """
    raise NotImplementedError



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""squad_question_generation dataset."""

import json
import os
import textwrap

from etils import epath
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.question_answering import qa_utils

_CITATION_DU_ET_AL = textwrap.dedent(
    """\
@inproceedings{du-etal-2017-learning,
    title = "Learning to Ask: Neural Question Generation for Reading Comprehension",
    author = "Du, Xinya  and Shao, Junru  and Cardie, Claire",
    booktitle = "Proceedings of the 55th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)",
    month = jul,
    year = "2017",
    address = "Vancouver, Canada",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/P17-1123",
    doi = "10.18653/v1/P17-1123",
    pages = "1342--1352",
}
"""
)

_DATA_URLS_DU_ET_AL = {
    "train": "https://raw.githubusercontent.com/xinyadu/nqg/master/data/raw/train.json",
    "dev": (
        "https://raw.githubusercontent.com/xinyadu/nqg/master/data/raw/dev.json"
    ),
    "test": "https://raw.githubusercontent.com/xinyadu/nqg/master/data/raw/test.json",
}

_HOMEPAGE_URL_DU_ET_AL = "https://github.com/xinyadu/nqg"

_CITATION_ZHOU_ET_AL = textwrap.dedent(
    """\
@inproceedings{du-etal-2017-learning,
    title = "Learning to Ask: Neural Question Generation for Reading Comprehension",
    author = "Du, Xinya  and Shao, Junru  and Cardie, Claire",
    booktitle = "Proceedings of the 55th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)",
    month = jul,
    year = "2017",
    address = "Vancouver, Canada",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/P17-1123",
    doi = "10.18653/v1/P17-1123",
    pages = "1342--1352",
}
"""
)

_DATA_URLS_ZHOU_ET_AL = {
    "train": (
        "https://rajpurkar.github.io/SQuAD-explorer/dataset/train-v1.1.json"
    ),
    "dev": "https://rajpurkar.github.io/SQuAD-explorer/dataset/dev-v1.1.json",
    "mapping": "https://res.qyzhou.me/qas_id_in_squad.zip",
    "redistribute": "https://res.qyzhou.me/redistribute.zip",
}

_HOMEPAGE_URL_ZHOU_ET_AL = "https://github.com/magic282/NQG"

_CITATION_SQUAD = textwrap.dedent(
    """\
@inproceedings{rajpurkar-etal-2016-squad,
    title = "{SQ}u{AD}: 100,000+ Questions for Machine Comprehension of Text",
    author = "Rajpurkar, Pranav  and Zhang, Jian  and Lopyrev, Konstantin  and Liang, Percy",
    booktitle = "Proceedings of the 2016 Conference on Empirical Methods in Natural Language Processing",
    month = nov,
    year = "2016",
    address = "Austin, Texas",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/D16-1264",
    doi = "10.18653/v1/D16-1264",
    pages = "2383--2392",
}
"""
)

_DESCRIPTION = """\
Question generation using squad dataset using data splits described in 'Neural
Question Generation from Text: A Preliminary Study' (Zhou et al, 2017) and
'Learning to Ask: Neural Question Generation for Reading Comprehension' (Du et
al, 2017).
"""

_CONTEXT_SENTENCE = "context_sentence"
_CONTEXT_PASSAGE = "context_passage"
_ANSWER = "answer"
_QUESTION = "question"


class SquadQuestionGenerationConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Squad Question Generation."""

  def __init__(self, *, features, data_urls, citation, **kwargs):
    """BuilderConfig for Squad Question Generation.

    Args:
      features: `tfds.features.FeaturesDict`, specific feature dict for the
        dataset.
      data_urls: `dict`, urls to download the files from
      citation: `string`, citation for the data set
      **kwargs: keyword arguments forwarded to super.
    """
    super(SquadQuestionGenerationConfig, self).__init__(**kwargs)
    self.features = features
    self.data_urls = data_urls
    self.citation = citation


class SquadQuestionGeneration(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for squad_question_generation dataset."""

  VERSION = tfds.core.Version("3.0.0")
  RELEASE_NOTES = {
      "1.0.0": """
          Initial build with unique SQuAD QAS ids in each split, using
          passage-level context (Zhou et al, 2017).
          """,
      "2.0.0": """
          Matches the original split of (Zhou et al, 2017), allows both
          sentence- and passage-level contexts, and uses answers from
          (Zhou et al, 2017).
          """,
      "3.0.0": """
          Added the split of (Du et al, 2017) also.
          """,
  }
  BUILDER_CONFIGS = [
      SquadQuestionGenerationConfig(
          name="split_du",
          description=textwrap.dedent(
              """\
          Answer independent question generation from passage-level contexts
          (Du et al, 2017).
          """
          ),
          features=tfds.features.FeaturesDict({
              _CONTEXT_PASSAGE: tfds.features.Text(),
              _ANSWER: tfds.features.Text(),
              _QUESTION: tfds.features.Text(),
          }),
          data_urls=_DATA_URLS_DU_ET_AL,
          citation=_CITATION_DU_ET_AL,
      ),
      SquadQuestionGenerationConfig(
          name="split_zhou",
          description=textwrap.dedent(
              """\
          Answer-span dependent question generation from sentence- and
          passage-level contexts (Zhou et al, 2017).
          """
          ),
          features=tfds.features.FeaturesDict({
              _CONTEXT_SENTENCE: tfds.features.Text(),
              _CONTEXT_PASSAGE: tfds.features.Text(),
              _ANSWER: tfds.features.Text(),
              _QUESTION: tfds.features.Text(),
          }),
          data_urls=_DATA_URLS_ZHOU_ET_AL,
          citation=_CITATION_ZHOU_ET_AL,
      ),
  ]

  def _info(self):
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(self.builder_config.features),
        homepage=_HOMEPAGE_URL_DU_ET_AL + "\n" + _CITATION_ZHOU_ET_AL,
        citation=self.builder_config.citation + "\n" + _CITATION_SQUAD,
        supervised_keys=(_CONTEXT_PASSAGE, _QUESTION),
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dl_paths = dl_manager.download_and_extract(self.builder_config.data_urls)

    if self.builder_config.name == "split_du":
      return {
          tfds.Split.TRAIN: self._generate_examples(dl_paths["train"]),
          tfds.Split.VALIDATION: self._generate_examples(dl_paths["dev"]),
          tfds.Split.TEST: self._generate_examples(dl_paths["test"]),
      }
    elif self.builder_config.name == "split_zhou":
      mapping_dir = os.path.join(dl_paths["mapping"], "qas_id_in_squad")
      redistribute_raw_dir = os.path.join(
          dl_paths["redistribute"], "redistribute/raw"
      )
      return {
          tfds.Split.TRAIN: self._generate_examples(
              dl_paths["train"],
              os.path.join(mapping_dir, "train.txt.id"),
              os.path.join(redistribute_raw_dir, "train.txt"),
          ),
          tfds.Split.VALIDATION: self._generate_examples(
              dl_paths["dev"],
              os.path.join(mapping_dir, "dev.txt.shuffle.dev.id"),
              os.path.join(redistribute_raw_dir, "dev.txt.shuffle.dev"),
          ),
          tfds.Split.TEST: self._generate_examples(
              dl_paths["dev"],
              os.path.join(mapping_dir, "dev.txt.shuffle.test.id"),
              os.path.join(redistribute_raw_dir, "dev.txt.shuffle.test"),
          ),
      }

  def _generate_examples(
      self, squad_data_path, mapping_path=None, qgen_data_path=None
  ):
    r"""Yields question generation examples.

    Args:
      squad_data_path: Path to SQuAD json file.
      mapping_path:  File with SQuAD map id for the example in Zhou et al.
        splits.
      qgen_data_path:  File with examples in "TokenizedInputSentence\t
        AnswerStartAndEndPosition\tParsingTreeOfInputSentence\t
        PoSTagOfInputSentence\tNERTagsOfInputSentence\tTokenizedQuestion\t
        UntokenizedInputSentence\tAnswerStartCharIndex\tAnswer\t
        UntokenizedQuestion" format per line, for the Zhou et al. splits.

    Yields:
      key and example dict.
    """
    if self.builder_config.name == "split_du":
      # The file format slightly differs from the original SQuAD JSON format.
      with epath.Path(squad_data_path).open() as f:
        squad_data = json.load(f)
        for article in squad_data:
          for paragraph in article["paragraphs"]:
            context = paragraph["context"]
            for qa in paragraph["qas"]:
              yield qa["id"], {
                  _CONTEXT_PASSAGE: context,
                  _ANSWER: qa["answers"][0]["text"],
                  _QUESTION: qa["question"],
              }
    elif self.builder_config.name == "split_zhou":
      squad_data = {}
      for k, ex in qa_utils.generate_squadlike_examples(squad_data_path):
        squad_data[k] = ex
      with tf.io.gfile.GFile(mapping_path, "r") as mapping_file:
        with tf.io.gfile.GFile(qgen_data_path, "r") as qgen_data_file:
          for ex_id, (squad_id, qgen_data) in enumerate(
              zip(
                  mapping_file.read().splitlines(),
                  qgen_data_file.read().splitlines(),
              )
          ):
            (_, _, _, _, _, _, context_sentence, _, answer, question) = (
                qgen_data.split("\t")
            )
            context_passage = squad_data[squad_id]["context"]
            yield str(ex_id).zfill(7), {
                _CONTEXT_SENTENCE: context_sentence,
                _CONTEXT_PASSAGE: context_passage,
                _ANSWER: answer,
                _QUESTION: question,
            }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wordnet dataset."""

import os

from etils import epath
import tensorflow_datasets.public_api as tfds

_LICENSE = """WordNet Release 3.0 This software and database is being provided
to you, the LICENSEE, by Princeton University under the following license. By
obtaining, using and/or copying this software and database, you agree that you
have read, understood, and will comply with these terms and conditions.:
Permission to use, copy, modify and distribute this software and database and
its documentation for any purpose and without fee or royalty is hereby granted,
provided that you agree to comply with the following copyright notice and
statements, including the disclaimer, and that the same appear on ALL copies of
the software, database and documentation, including modifications that you make
for internal use or for distribution. WordNet 3.0 Copyright 2006 by Princeton
University. All rights reserved. THIS SOFTWARE AND DATABASE IS PROVIDED "AS IS"
AND PRINCETON UNIVERSITY MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
IMPLIED. BY WAY OF EXAMPLE, BUT NOT LIMITATION, PRINCETON UNIVERSITY MAKES NO
REPRESENTATIONS OR WARRANTIES OF MERCHANT- ABILITY OR FITNESS FOR ANY PARTICULAR
PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE, DATABASE OR DOCUMENTATION WILL
NOT INFRINGE ANY THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS.
The name of Princeton University or Princeton may not be used in advertising or
publicity pertaining to distribution of the software and/or database. Title to
copyright in this software, database and any associated documentation shall at
all times remain with Princeton University and LICENSEE agrees to preserve same.
"""

_CITATION = """@article{10.1145/219717.219748,
author = {Miller, George A.},
title = {WordNet: A Lexical Database for English},
year = {1995},
issue_date = {Nov. 1995},
publisher = {Association for Computing Machinery},
address = {New York, NY, USA},
volume = {38},
number = {11},
issn = {0001-0782},
url = {https://doi.org/10.1145/219717.219748},
doi = {10.1145/219717.219748},
journal = {Commun. ACM},
month = nov,
pages = {39--41},
numpages = {3}
}
"""

_DESCRIPTION = """WordNet is a large lexical database of English. Nouns, verbs,
adjectives and adverbs are grouped into sets of cognitive synonyms (synsets),
each expressing a distinct concept. Synsets are interlinked by means of
conceptual-semantic and lexical relations.
"""

_WN18_DESCRIPTION = """This WORDNET TENSOR DATA consists of a collection of
triplets (synset, relation_type, triplet) extracted from WordNet 3.0
(http://wordnet.princeton.edu). This data set can be seen as a 3-mode tensor
depicting ternary relationships between synsets. See
https://everest.hds.utc.fr/doku.php?id=en:transe.
"""

_WN18_CITATION = """@incollection{NIPS2013_5071,
title = {Translating Embeddings for Modeling Multi-relational Data},
author = {Bordes, Antoine and Usunier, Nicolas and Garcia-Duran, Alberto and Weston, Jason and Yakhnenko, Oksana},
booktitle = {Advances in Neural Information Processing Systems 26},
editor = {C. J. C. Burges and L. Bottou and M. Welling and Z. Ghahramani and K. Q. Weinberger},
pages = {2787--2795},
year = {2013},
publisher = {Curran Associates, Inc.},
url = {http://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data.pdf}
}
"""

_WN18RR_DESCRIPTION = """Same as WN18 but fixes test leakage through inverse
relations. See https://github.com/TimDettmers/ConvE.
"""

_WN18RR_CITATION = """@inproceedings{dettmers2018conve,
	Author = {Dettmers, Tim and Pasquale, Minervini and Pontus, Stenetorp and Riedel, Sebastian},
	Booktitle = {Proceedings of the 32th AAAI Conference on Artificial Intelligence},
	Title = {Convolutional 2D Knowledge Graph Embeddings},
	Url = {https://arxiv.org/abs/1707.01476},
	Year = {2018},
        pages  = {1811--1818},
  	Month = {February}
}
"""

_RELATIONS = [
    '_also_see',
    '_derivationally_related_form',
    '_has_part',
    '_hypernym',
    '_instance_hypernym',
    '_member_meronym',
    '_member_of_domain_region',
    '_member_of_domain_usage',
    '_similar_to',
    '_synset_domain_topic_of',
    '_verb_group',
]


def _make_wn18_metadata(synset_definitions_path):
  synsets = {}
  with epath.Path(synset_definitions_path).open() as f:
    for line in f:
      synset_id, name, definition = line.strip().split('\t')
      synsets[synset_id] = dict(name=name, definition=definition)
  return dict(relations=_RELATIONS, synsets=synsets)


class WordnetConfig(tfds.core.BuilderConfig):
  """Configuration for `Wordnet`."""

  def __init__(self, name, path_prefix, description, citation, version):
    self._citation = citation
    self._path_prefix = path_prefix
    super(WordnetConfig, self).__init__(
        name=name, description=description, version=version
    )

  @property
  def citation(self):
    return '\n'.join([_CITATION, self._citation])

  def get_paths(self, dl_paths):
    root_dir = dl_paths[self.name]
    return (
        os.path.join(root_dir, self._path_prefix + 'train.txt'),
        os.path.join(root_dir, self._path_prefix + 'valid.txt'),
        os.path.join(root_dir, self._path_prefix + 'test.txt'),
    )


class Wordnet(tfds.core.GeneratorBasedBuilder):
  """Builder for WordNet dataset."""

  BUILDER_CONFIGS = [
      WordnetConfig(
          name='WN18',
          path_prefix=os.path.join('wordnet-mlj12', 'wordnet-mlj12-'),
          description=_WN18_DESCRIPTION,
          citation=_WN18_CITATION,
          version=tfds.core.Version('0.1.0'),
      ),
      WordnetConfig(
          name='WN18RR',
          path_prefix='',
          description=_WN18RR_DESCRIPTION,
          citation=_WN18RR_CITATION,
          version=tfds.core.Version('0.1.0'),
      ),
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'lhs': tfds.features.Text(),
            'relation': tfds.features.Text(),
            'rhs': tfds.features.Text(),
        }),
        homepage='https://wordnet.princeton.edu/',
        citation=self.builder_config.citation,
        metadata=tfds.core.MetadataDict(),
        license=_LICENSE,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dl_paths = dl_manager.download_and_extract({
        'WN18': 'https://everest.hds.utc.fr/lib/exe/fetch.php?media=en:wordnet-mlj12.tar.gz',
        'WN18RR': (
            'https://github.com/TimDettmers/ConvE/raw/master/WN18RR.tar.gz'
        ),
    })
    # Metadata is at the configuration level and is the same for all splits.
    synset_definitions_path = os.path.join(
        dl_paths['WN18'], 'wordnet-mlj12', 'wordnet-mlj12-definitions.txt'
    )
    self.info.metadata.update(_make_wn18_metadata(synset_definitions_path))
    # Locate and output splits.
    train_path, val_path, test_path = self.builder_config.get_paths(dl_paths)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs=dict(triplets_path=train_path),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs=dict(triplets_path=val_path),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs=dict(triplets_path=test_path),
        ),
    ]

  def _generate_examples(self, triplets_path):
    """Yields examples."""
    with epath.Path(triplets_path).open() as f:
      for i, line in enumerate(f):
        lhs, relation, rhs = line.strip().split('\t')
        yield i, {'lhs': lhs, 'relation': relation, 'rhs': rhs}



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""mrqa datasets."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text.mrqa import mrqa


class MRQASQuADTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for squad dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['squad']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'train': 9,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
  }

  DL_EXTRACT_RESULT = {
      'train': 'squad/train.jsonl',
      'validation': 'squad/validation.jsonl',
  }


class MRQANewsQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for news_qa dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['news_qa']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'train': 9,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
  }

  DL_EXTRACT_RESULT = {
      'train': 'news_qa/train.jsonl',
      'validation': 'news_qa/validation.jsonl',
  }


class MRQATriviaQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for trivia_qa dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['trivia_qa']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'train': 9,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
  }

  DL_EXTRACT_RESULT = {
      'train': 'trivia_qa/train.jsonl',
      'validation': 'trivia_qa/validation.jsonl',
  }


class MRQASearchQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for search_qa dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['search_qa']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'train': 9,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
  }

  DL_EXTRACT_RESULT = {
      'train': 'search_qa/train.jsonl',
      'validation': 'search_qa/validation.jsonl',
  }


class MRQAHotpotQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for hotpot_qa dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['hotpot_qa']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'train': 9,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
  }

  DL_EXTRACT_RESULT = {
      'train': 'hotpot_qa/train.jsonl',
      'validation': 'hotpot_qa/validation.jsonl',
  }


class MRQANaturalQuestionsTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for natural_questions dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['natural_questions']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'train': 9,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
  }

  DL_EXTRACT_RESULT = {
      'train': 'natural_questions/train.jsonl',
      'validation': 'natural_questions/validation.jsonl',
  }


class MRQABioASQTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for bio_asq dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['bio_asq']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'test': 1,  # Number of fake test examples.
  }

  DL_EXTRACT_RESULT = {
      'test': 'bio_asq/test.jsonl',
  }


class MRQADROPTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for drop dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['drop']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'test': 1,  # Number of fake test examples.
  }

  DL_EXTRACT_RESULT = {
      'test': 'drop/test.jsonl',
  }


class MRQADuoRCTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for duo_rc dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['duo_rc']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'test': 1,  # Number of fake test examples.
  }

  DL_EXTRACT_RESULT = {
      'test': 'duo_rc/test.jsonl',
  }


class MRQARACETest(tfds.testing.DatasetBuilderTestCase):
  """Tests for race dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['race']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'test': 1,  # Number of fake test examples.
  }

  DL_EXTRACT_RESULT = {
      'test': 'race/test.jsonl',
  }


class MRQARelationExtractionTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for relation_extraction dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['relation_extraction']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'test': 1,  # Number of fake test examples.
  }

  DL_EXTRACT_RESULT = {
      'test': 'relation_extraction/test.jsonl',
  }


class MRQATextbookQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for textbook_qa dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['textbook_qa']
  DATASET_CLASS = mrqa.MRQA
  SPLITS = {
      'test': 1,  # Number of fake test examples.
  }

  DL_EXTRACT_RESULT = {
      'test': 'textbook_qa/test.jsonl',
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""gem dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text.gem import gem


class GemCommonGenTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['common_gen']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_scramble': 1,
  }

  DL_EXTRACT_RESULT = {'data': 'common_gen', 'challenge_set': 'challenge_sets'}


class GemCsRestaurantsTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['cs_restaurants']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_scramble': 1,
  }

  DL_EXTRACT_RESULT = {
      'train': 'cs_restaurants/train.jsonl',
      'validation': 'cs_restaurants/devel.jsonl',
      'test': 'cs_restaurants/test.jsonl',
      'challenge_set': 'challenge_sets',
  }


class GemDartTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['dart']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
  }

  DL_EXTRACT_RESULT = {
      'train': 'dart/dart-v1.1.1-full-train.json',
      'validation': 'dart/dart-v1.1.1-full-dev.json',
      'test': 'dart/dart-v1.1.1-full-test.json',
  }


class GemE2ETest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['e2e_nlg']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_scramble': 1,
  }

  DL_EXTRACT_RESULT = {
      'train': 'e2e_nlg/train-fixed.no-ol.csv',
      'validation': 'e2e_nlg/devel-fixed.no-ol.csv',
      'test': 'e2e_nlg/test-fixed.csv',
      'challenge_set': 'challenge_sets',
  }


class GemMLSumTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['mlsum_de', 'mlsum_es']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_covid': 1,
  }

  DL_EXTRACT_RESULT = {
      'train': 'mlsum',
      'validation': 'mlsum',
      'test': 'mlsum',
      'bad_ids': 'mlsum/gem_mlsum_bad_ids.json',
      'challenge_set': 'challenge_sets',
  }


class GemSchemaGuidedDialogTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['schema_guided_dialog']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_backtranslation': 1,
      'challenge_test_bfp02': 1,
      'challenge_test_bfp05': 1,
      'challenge_test_nopunc': 1,
      'challenge_test_scramble': 1,
  }

  DL_EXTRACT_RESULT = {
      'data': 'schema_guided_dialog',
      'challenge_set': 'challenge_sets',
  }


class GemTottoTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['totto']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_scramble': 1,
  }

  DL_EXTRACT_RESULT = {'data': '', 'challenge_set': 'challenge_sets'}


class GemWebNlgEnTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['web_nlg_en']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_scramble': 1,
      'challenge_test_numbers': 1,
  }

  DL_EXTRACT_RESULT = {
      'train': 'web_nlg/webnlg_en_train.json',
      'validation': 'web_nlg/webnlg_en_val.json',
      'test': 'web_nlg/webnlg_en_test.json',
      'challenge_set': 'challenge_sets',
  }


class GemWebNlgRuTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['web_nlg_ru']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_scramble': 1,
  }

  DL_EXTRACT_RESULT = {
      'train': 'web_nlg/webnlg_ru_train.json',
      'validation': 'web_nlg/webnlg_ru_val.json',
      'test': 'web_nlg/webnlg_ru_test.json',
      'challenge_set': 'challenge_sets',
  }


class GemWikiAutoTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['wiki_auto_asset_turk']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test_turk': 1,  # Number of fake test examples.
      'test_asset': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_asset_backtranslation': 1,
      'challenge_test_asset_bfp02': 1,
      'challenge_test_asset_bfp05': 1,
      'challenge_test_asset_nopunc': 1,
      'challenge_test_turk_backtranslation': 1,
      'challenge_test_turk_bfp02': 1,
      'challenge_test_turk_bfp05': 1,
      'challenge_test_turk_nopunc': 1,
  }

  DL_EXTRACT_RESULT = {
      'train': 'wiki_auto_asset_turk/train.tsv',
      'validation': 'wiki_auto_asset_turk/valid.tsv',
      'test_asset_0': 'wiki_auto_asset_turk/asset.test.simp.0',
      'test_asset_1': 'wiki_auto_asset_turk/asset.test.simp.1',
      'test_asset_2': 'wiki_auto_asset_turk/asset.test.simp.2',
      'test_asset_3': 'wiki_auto_asset_turk/asset.test.simp.3',
      'test_asset_4': 'wiki_auto_asset_turk/asset.test.simp.4',
      'test_asset_5': 'wiki_auto_asset_turk/asset.test.simp.5',
      'test_asset_6': 'wiki_auto_asset_turk/asset.test.simp.6',
      'test_asset_7': 'wiki_auto_asset_turk/asset.test.simp.7',
      'test_asset_8': 'wiki_auto_asset_turk/asset.test.simp.8',
      'test_asset_9': 'wiki_auto_asset_turk/asset.test.simp.9',
      'test_turk': 'wiki_auto_asset_turk/gem_turk_detokenized.json',
      'challenge_set': 'challenge_sets',
  }


class GemWikiLinguaTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = [
      'wiki_lingua_arabic_ar',
      'wiki_lingua_chinese_zh',
      'wiki_lingua_czech_cs',
      'wiki_lingua_dutch_nl',
      'wiki_lingua_english_en',
      'wiki_lingua_french_fr',
      'wiki_lingua_german_de',
      'wiki_lingua_hindi_hi',
      'wiki_lingua_indonesian_id',
      'wiki_lingua_italian_it',
      'wiki_lingua_japanese_ja',
      'wiki_lingua_korean_ko',
      'wiki_lingua_portuguese_pt',
      'wiki_lingua_russian_ru',
      'wiki_lingua_spanish_es',
      'wiki_lingua_thai_th',
      'wiki_lingua_turkish_tr',
      'wiki_lingua_vietnamese_vi',
  ]
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
  }

  DL_EXTRACT_RESULT = {
      'data': 'wiki_lingua',
  }


class GemXsumTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gem dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['xsum']
  DATASET_CLASS = gem.Gem
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake dev examples.
      'test': 1,  # Number of fake test examples.
      'challenge_train_sample': 1,
      'challenge_validation_sample': 1,
      'challenge_test_backtranslation': 1,
      'challenge_test_bfp_02': 1,
      'challenge_test_bfp_05': 1,
      'challenge_test_nopunc': 1,
      'challenge_test_covid': 1,
  }

  DL_EXTRACT_RESULT = {
      'data': 'xsum',
      'splits': 'xsum/gem_xsum_confidence_0.8.json',
      'challenge_set': 'challenge_sets',
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ArrayRecord DataSource base class.

Warning: this is an experimental module. The interface might change in the
future without backwards compatibility.
"""

from collections.abc import Sequence as AbcSequence
import dataclasses
from typing import Any, Optional, Sequence, TypeVar

from absl import logging
from tensorflow_datasets.core import dataset_info as dataset_info_lib
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import array_record_data_source
import tree

T = TypeVar('T')

_DEFAULT_ITERATION_STEP = 1000


@dataclasses.dataclass
class ArrayRecordDataSource(AbcSequence):
  """Grain DataSource class.

  Warning: this is an experimental class. The interface might change in the
  future without backwards compatibility.

  It acts as a wrapper around `array_record.ArrayRecordDataSource` that can read
  from ArrayRecords. It exposes `__len__` and `__getitem__` to serve as a data
  source.
  """

  dataset_info: dataset_info_lib.DatasetInfo
  split: splits_lib.Split = None
  decoders: Optional[type_utils.TreeDict[decode.partial_decode.DecoderArg]] = (
      None
  )
  # In order to lazy load array_record, we don't load
  # `array_record_data_source.ArrayRecordDataSource` here.
  data_source: Any = dataclasses.field(init=False)
  length: int = dataclasses.field(init=False)

  def __post_init__(self):
    file_format = self.dataset_info.file_format
    if file_format != file_adapters.FileFormat.ARRAY_RECORD:
      raise NotImplementedError(
          f'No random access data source for file format {file_format}. Please,'
          ' generate your data using `tfds.builder(...,'
          f' file_format={file_adapters.FileFormat.ARRAY_RECORD})`.'
      )
    split_infos = self.dataset_info.splits.values()
    splits_dict = splits_lib.SplitDict(split_infos=split_infos)
    file_instructions = splits_dict[self.split].file_instructions
    self.data_source = array_record_data_source.ArrayRecordDataSource(
        file_instructions
    )
    self.length = len(self.data_source)

  def __len__(self) -> int:
    return self.length

  def __iter__(self):
    for i in range(self.length):
      yield self[i]

  def __getitem__(self, record_key: int) -> Any:
    if not isinstance(record_key, int):
      logging.error(
          'Calling ArrayRecordDataSource.__getitem__() with sequence '
          'of record keys (%s) is deprecated. Either pass a single '
          'integer or switch to __getitems__().',
          record_key,
      )
      return self.__getitems__(record_key)
    record = self.data_source[record_key]
    return self.dataset_info.features.deserialize_example_np(
        record, decoders=self.decoders
    )

  def __getitems__(self, record_keys: Sequence[int]) -> Sequence[Any]:
    """Retrieves items by batch.

    This method allows PyTorch to load records by batch, rather than one by one.

    Args:
      record_keys: a sequence of keys.

    Returns:
      The records associated with the keys.

    Raises:
      IndexError: If the number of retrieved records is incorrect.
    """
    records = self.data_source.__getitems__(record_keys)
    features = self.dataset_info.features
    if len(record_keys) != len(records):
      raise IndexError(
          f'Requested {len(record_keys)} records but got'
          f' {len(records)} records.'
          f'{record_keys=}, {records=}'
      )
    return [
        features.deserialize_example_np(record, decoders=self.decoders)
        for record in records
    ]

  def __repr__(self) -> str:
    decoders_repr = (
        tree.map_structure(type, self.decoders) if self.decoders else None
    )
    return (
        f'DataSource(name={self.dataset_info.name}, '
        f'split={self.split!r}, '
        f'decoders={decoders_repr})'
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mock util for tfds."""

from __future__ import annotations

import contextlib
import enum
import functools
import os
from typing import Any, Callable, Iterator, Optional, Sequence
from unittest import mock

from absl import logging
from etils import epath
import numpy as np
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import read_only_builder
from tensorflow_datasets.core import reader as reader_lib
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.data_sources import array_record
from tensorflow_datasets.core.utils import dtype_utils
from tensorflow_datasets.core.utils import tree_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.testing import test_utils


def _get_fake_data_components(decoders, features):
  """Gets all the components to generate fake data in the tests.

  Args:
    decoders: The decoders to override, or `None` if no decoding is used.
    features: The original features.

  Returns:
    A tuple with the data generator class, the features, the feature specs and
      the decode function.
  """
  # Partial decoding
  if isinstance(decoders, decode.PartialDecoding):
    # TODO(epot): Should be moved inside `features.decode_example`
    features = decoders.extract_features(features)
    decoders = decoders.decoders
  # Full decoding (all features decoded)
  else:
    decoders = decoders  # pylint: disable=self-assigning-variable

  has_nested_dataset = any(
      isinstance(f, features_lib.Dataset) for f in features._flatten(features)  # pylint: disable=protected-access
  )
  if decoders is not None or has_nested_dataset:
    # If a decoder is passed, encode/decode the examples.
    generator_cls = EncodedRandomFakeGenerator
    specs = features.get_serialized_info()
    decode_fn = functools.partial(features.decode_example, decoders=decoders)
  else:
    generator_cls = RandomFakeGenerator
    specs = features.get_tensor_info()
    decode_fn = lambda ex: ex  # identity
  return generator_cls, features, specs, decode_fn


class PickableDataSourceMock(mock.MagicMock):
  """Makes MagicMock pickable in order to work with multiprocessing in Grain."""

  def __getstate__(self):
    return {'num_examples': len(self), 'generator': self._generator}

  def __setstate__(self, state):
    num_examples, generator = state['num_examples'], state['generator']
    self.__len__.return_value = num_examples
    self.__getitem__ = functools.partial(_getitem, generator=generator)
    self.__getitems__ = functools.partial(_getitems, generator=generator)

  def __reduce__(self):
    return (PickableDataSourceMock, (), self.__getstate__())


def _getitem(self, record_key: int, generator: RandomFakeGenerator) -> Any:
  """Function to overwrite __getitem__ in data sources."""
  del self
  return generator[record_key]


def _getitems(
    self, record_keys: Sequence[int], generator: RandomFakeGenerator
) -> Sequence[Any]:
  """Function to overwrite __getitems__ in data sources."""
  del self
  return [generator[record_key] for record_key in record_keys]


def _deserialize_example_np(serialized_example, *, decoders=None):
  """Function to overwrite dataset_info.features.deserialize_example_np.

  Warning: this has to be defined in the outer scope in order for the function
  to be pickable.

  Args:
    serialized_example: the example to deserialize.
    decoders: optional decoders.

  Returns:
    The serialized example, because deserialization is taken care by
      RandomFakeGenerator.
  """
  del decoders
  return serialized_example


class MockPolicy(enum.Enum):
  """Strategy to use with `tfds.testing.mock_data` to mock the dataset.

  Attributes:
    AUTO: Try `USE_FILES`, and fallback to `USE_CODE` if the metadata files
      (`dataset_info.json`,...) are not found.
    USE_FILES: Load dataset from the metadata files (present in the `data_dir`
      kwarg of `tfds.testing.mock_data`, raise error if data_dir is not
      reachable.
    USE_CODE: Load the data from the original dataset generation class. Do not
      use any generated files. More is more convenient but less safe than
      `USE_FILES`. Not all features might be available (e.g. no split-names).
  """

  AUTO = enum.auto()
  USE_CODE = enum.auto()
  USE_FILES = enum.auto()


@contextlib.contextmanager
def mock_data(
    num_examples: int = 1,
    num_sub_examples: int = 1,
    max_value: Optional[int] = None,
    *,
    policy: MockPolicy = MockPolicy.AUTO,
    as_dataset_fn: Optional[Callable[..., tf.data.Dataset]] = None,
    data_dir: Optional[str] = None,
    mock_array_record_data_source: Optional[PickableDataSourceMock] = None,
) -> Iterator[None]:
  """Mock tfds to generate random data.

  ### Usage

  * Usage (automated):

  ```py
  with tfds.testing.mock_data(num_examples=5):
    ds = tfds.load('some_dataset', split='train')

    for ex in ds:  # ds will yield randomly generated examples.
      ex
  ```

  All calls to `tfds.load`/`tfds.data_source` within the context manager then
  return deterministic mocked data.

  * Usage (manual):

  For more control over the generated examples, you can
  manually overwrite the `DatasetBuilder._as_dataset` method:

  ```py
  def as_dataset(self, *args, **kwargs):
    return tf.data.Dataset.from_generator(
        lambda: ({
            'image': np.ones(shape=(28, 28, 1), dtype=np.uint8),
            'label': i % 10,
        } for i in range(num_examples)),
        output_types=self.info.features.dtype,
        output_shapes=self.info.features.shape,
    )

  with mock_data(as_dataset_fn=as_dataset):
    ds = tfds.load('some_dataset', split='train')

    for ex in ds:  # ds will yield the fake data example of 'as_dataset'.
      ex
  ```

  ### Policy

  For improved results, you can copy the true metadata files
  (`dataset_info.json`, `label.txt`, vocabulary files) in
  `data_dir/dataset_name/version`. This will allow the mocked dataset to use
  the true metadata computed during generation (split names,...).

  If metadata files are not found, then info from the original class will be
  used, but the features computed during generation won't be available (e.g.
  unknown split names, so any splits are accepted).

  ### Miscellaneous

  * The examples are deterministically generated. Train and test split will
    yield the same examples.
  * The actual examples will be randomly generated using
    `builder.info.features.get_tensor_info()`.
  * Download and prepare step will always be a no-op.
  * Warning: `info.split['train'].num_examples` won't match
    `len(list(ds_train))`

  Some of those points could be improved. If you have suggestions, issues with
  this functions, please open a new issue on our Github.

  Args:
    num_examples: Number of fake example to generate.
    num_sub_examples: Number of examples to generate in nested Dataset features.
    max_value: The maximum value present in generated tensors; if max_value is
      None or it is set to 0, then random numbers are generated from the range
      from 0 to 255.
    policy: Strategy to use to generate the fake examples. See
      `tfds.testing.MockPolicy`.
    as_dataset_fn: If provided, will replace the default random example
      generator. This function mock the `FileAdapterBuilder._as_dataset`
    data_dir: Folder containing the metadata file (searched in
      `data_dir/dataset_name/version`). Overwrite `data_dir` kwargs from
      `tfds.load`. Used in `MockPolicy.USE_FILES` mode.
    mock_array_record_data_source: Overwrite a mock for the underlying
      ArrayRecord data source if it is used.

  Yields:
    None
  """

  original_init_fn = dataset_builder.DatasetBuilder.__init__
  original_as_dataset_fn = dataset_builder.DatasetBuilder.as_dataset
  original_builder_from_files = read_only_builder.builder_from_files
  if mock_array_record_data_source is None:
    mock_array_record_data_source = PickableDataSourceMock()
  elif not isinstance(mock_array_record_data_source, PickableDataSourceMock):
    raise ValueError(
        '`mock_array_record_data_source` must be a'
        ' `tfds.testing.PickableDataSourceMock` because it must be pickable.'
    )

  def mock_download_and_prepare(self, *args, **kwargs):
    """`builder.download_and_prepare` is a no-op."""
    del self, args, kwargs  # Unused

  def _check_policy(self) -> MockPolicy:
    """Checks the policy and returns the actual used policy for mocking."""
    # When `USE_FILES` is used, make sure the metadata actually exists.
    if epath.Path(self.data_dir).exists():
      logging.info('Metadata found for %s at %s', self.name, self.data_dir)
    else:
      if policy == MockPolicy.USE_FILES:
        raise ValueError(
            'TFDS has been mocked with `MockPolicy.USE_FILES`, but metadata '
            f'files were not found in {self.data_dir}. '
            'You should copy the real metadata files, so that the dataset '
            'can be loaded properly, or set the data_dir kwarg of '
            'tfds.testing.mock_tfds(data_dir=...).\n'
        )
      if policy == MockPolicy.AUTO:
        logging.info(
            'Metadata NOT found for %s at %s. Will use `MockPolicy.USE_CODE.`',
            self.name,
            self.data_dir,
        )
        return MockPolicy.USE_CODE
    return policy

  def mock_as_dataset_base(self, **kwargs):
    """Mocks `builder.as_dataset`."""
    _check_policy(self)

    # Info is already restored at this point, so can mock the file system
    # safely in case of `USE_CODE` mode.
    # The only gfile access is to check `self.data_dir` existance.
    with test_utils.MockFs() as fs:
      fs.add_file(os.path.join(self.data_dir, 'tmp.txt'))
      return original_as_dataset_fn(self, **kwargs)

  def mock_as_dataset(self, split, decoders=None, read_config=None, **kwargs):
    """Mocks `builder._as_dataset`."""
    del kwargs

    generator_cls, features, specs, decode_fn = _get_fake_data_components(
        decoders, self.info.features
    )

    ds = tf.data.Dataset.from_generator(
        # `from_generator` takes a callable with signature () -> iterable
        # Recreating a new generator each time ensure that all pipelines are
        # using the same examples
        functools.partial(
            generator_cls,
            features=features,
            num_examples=num_examples,
            num_sub_examples=num_sub_examples,
            max_value=max_value,
        ),
        # pylint: enable=g-long-lambda]
        output_types=tf.nest.map_structure(lambda t: t.dtype, specs),
        output_shapes=tf.nest.map_structure(lambda t: t.shape, specs),
    )
    ds = ds.apply(tf.data.experimental.assert_cardinality(num_examples))
    ds = ds.map(decode_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)

    if read_config and read_config.add_tfds_id:
      ds_id = reader_lib._make_id_dataset(  # pylint: disable=protected-access
          filename=f'{self.name}-{split}.tfrecord-00000-of-00001',
          start_index=0,  # pytype: disable=wrong-arg-types
      )
      ds = tf.data.Dataset.zip((ds, ds_id))
      ds = ds.map(lambda ex, id: {'tfds_id': id, **ex})

    return ds

  def mock_as_data_source(self, split, decoders=None, **kwargs):
    """Mocks `builder.as_data_source`."""
    del kwargs
    actual_policy = _check_policy(self)
    if split is None:
      split = {s: s for s in self.info.splits}

    generator_cls, features, _, _ = _get_fake_data_components(
        decoders, self.info.features
    )
    generator = generator_cls(features, num_examples)

    if actual_policy == MockPolicy.USE_CODE:
      # Mock `SplitDict`, because splits are not known in advance.
      split_dict = mock.MagicMock(spec=splits_lib.SplitDict)
    else:
      # Use the real `SplitDict`.
      split_dict = splits_lib.SplitDict
    with mock.patch.object(
        splits_lib,
        'SplitDict',
        new=split_dict,
    ), mock.patch(
        'array_record.python.array_record_data_source.ArrayRecordDataSource',
        mock_array_record_data_source,
    ), mock.patch(
        'tensorflow_datasets.core.dataset_info.DatasetInfo.file_format',
        new_callable=mock.PropertyMock,
        # Force ARRAY_RECORD as the default file_format.
        return_value=file_adapters.FileFormat.ARRAY_RECORD,
    ):
      self.info.features.deserialize_example_np = _deserialize_example_np
      mock_array_record_data_source.return_value.__len__.return_value = (
          num_examples
      )
      mock_array_record_data_source.return_value._generator = (  # pylint:disable=protected-access
          generator
      )
      mock_array_record_data_source.return_value.__getitem__ = (
          functools.partial(_getitem, generator=generator)
      )
      mock_array_record_data_source.return_value.__getitems__ = (
          functools.partial(_getitems, generator=generator)
      )

      def build_single_data_source(split):
        single_data_source = array_record.ArrayRecordDataSource(
            dataset_info=self.info, split=split, decoders=decoders
        )
        return single_data_source

      return tree_utils.map_structure(build_single_data_source, split)

  if not as_dataset_fn:
    as_dataset_fn = mock_as_dataset

  if policy == MockPolicy.USE_CODE:
    mock_data_dir = '/tmp/non-existing/tfds/data_dir/'
  elif not data_dir:
    mock_data_dir = os.path.join(os.path.dirname(__file__), 'metadata')
  else:  # AUTO or USE_FILES with explicitly given `data_dir`
    mock_data_dir = data_dir

  def mock_init(*args, data_dir=None, **kwargs):
    del data_dir  # Unused. Inject `mock_data_dir` instead.
    return original_init_fn(*args, data_dir=mock_data_dir, **kwargs)

  def new_builder_from_files(*args, **kwargs):
    # Replace the user-given data dir by the mocked one
    kwargs.pop('data_dir', None)
    # `DatasetBuilder.__init__` is mocked above to inject the wrong data_dir.
    # So we restore the original `DatasetBuilder.__init__` inside
    # `builder_from_files` calls.
    with mock.patch(
        f'{core}.dataset_builder.DatasetBuilder.__init__', original_init_fn
    ):
      return original_builder_from_files(
          *args, data_dir=mock_data_dir, **kwargs
      )

  core = 'tensorflow_datasets.core'
  with contextlib.ExitStack() as stack:
    for path, mocked_fn in [
        # All GCS access should fail (is_dataset_on_gcs,...).
        (f'{core}.utils.gcs_utils.exists', lambda path: False),
        # Patch `data_dir`: `data_dir` explicitly set by users will be ignored.
        # `data_dir` is used at two places:
        # * `builder_from_files` to search read-only datasets loaded from config
        # * `DatasetBuilder.__init__` otherwise
        (f'{core}.dataset_builder.DatasetBuilder.__init__', mock_init),
        (
            f'{core}.read_only_builder.builder_from_files',
            new_builder_from_files,
        ),
        # Patch DatasetBuilder
        (
            f'{core}.dataset_builder.DatasetBuilder.download_and_prepare',
            mock_download_and_prepare,
        ),
        (
            f'{core}.dataset_builder.DatasetBuilder.as_dataset',
            mock_as_dataset_base,
        ),
        (
            f'{core}.dataset_builder.DatasetBuilder.as_data_source',
            mock_as_data_source,
        ),
        (
            f'{core}.dataset_builder.FileReaderBuilder._as_dataset',
            as_dataset_fn,
        ),
    ]:
      stack.enter_context(mock.patch(path, mocked_fn))
    yield


class RandomFakeGenerator(object):
  """Generator of fake examples randomly and deterministically generated."""

  def __init__(
      self,
      features,
      num_examples: int,
      num_sub_examples: int = 1,
      max_value: Optional[int] = None,
      constant_dynamic_shape: bool = False,
      seed: int = 0,
  ):
    self._seed = seed
    self._features = features
    self._num_examples = num_examples
    self._num_sub_examples = num_sub_examples
    self._max_value = max_value
    self._constant_dynamic_shape = constant_dynamic_shape

  def _generate_random_string_array(self, shape, rng):
    """Generates an array of random strings."""

    def rand_str():
      return ''.join(
          rng.choice(list(' abcdefghij'), size=(rng.integers(low=10, high=20)))
      )

    if not shape:
      return rand_str()

    return np.array(
        [rand_str() for _ in range(np.prod(shape, dtype=np.int32))]
    ).reshape(shape)

  def _generate_random_obj(self, feature, tensor_info, rng):
    """Generates a random tensor for a single feature."""
    # TODO(tfds): Could improve the fake generatiion:
    # * Use the feature statistics (min, max)
    # * For Sequence features
    # * For Text

    # First we deal with the case of sub-datasets:
    if isinstance(feature, features_lib.Dataset):
      # For sub-datasets self._num_sub_examples examples are generated.
      generator = RandomFakeGenerator(
          feature.feature,
          num_examples=self._num_sub_examples,
          num_sub_examples=1,
          max_value=self._max_value,
      )
      # Returns the list of examples in the nested dataset.
      return list(generator)

    if self._constant_dynamic_shape:
      shape = [10 if s is None else s for s in tensor_info.shape]
    else:
      shape = [  # Fill dynamic shape with random values
          rng.integers(low=5, high=50) if s is None else s
          for s in tensor_info.shape
      ]
    if isinstance(feature, features_lib.ClassLabel):
      max_value = feature.num_classes
    elif isinstance(feature, features_lib.Text) and feature.vocab_size:
      max_value = feature.vocab_size
    elif self._max_value:
      max_value = self._max_value
    else:
      max_value = 255

    # We cast the data to make sure `encode_example` don't raise errors
    dtype = tensor_info.np_dtype
    # Generate some random values, depending on the dtype
    if dtype_utils.is_integer(dtype):
      return rng.integers(0, max_value, size=shape).astype(dtype)
    elif dtype_utils.is_floating(dtype):
      return rng.random(size=shape).astype(dtype)
    elif dtype_utils.is_bool(dtype):
      return (rng.random(size=shape) < 0.5).astype(dtype)
    elif dtype_utils.is_string(dtype):
      return self._generate_random_string_array(shape, rng)
    raise ValueError('Fake generation not supported for {}'.format(dtype))

  def _generate_example(self, rng):
    """Generates the next example."""
    root_feature = self._features
    flat_features = root_feature._flatten(root_feature)  # pylint: disable=protected-access
    flat_tensor_info = root_feature._flatten(root_feature.get_tensor_info())  # pylint: disable=protected-access
    flat_objs = [
        self._generate_random_obj(feature, tensor_info, rng)
        for feature, tensor_info in zip(flat_features, flat_tensor_info)
    ]
    return root_feature._nest(flat_objs)  # pylint: disable=protected-access

  def __iter__(self):
    """Yields all fake examples deterministically."""
    rng = np.random.default_rng(self._seed)
    for _ in range(self._num_examples):
      yield self._generate_example(rng)

  def __getitem__(self, index: int):
    """Returns a single fake example deterministically."""
    rng = np.random.default_rng(self._seed + index)
    return self._generate_example(rng)


class EncodedRandomFakeGenerator(RandomFakeGenerator):
  """Generator of fake encoded examples."""

  def __iter__(self):
    """Yields all fake examples."""
    for ex in super().__iter__():
      yield self._features.encode_example(ex)

  def __getitem__(self, index: int):
    decoded_example = super().__getitem__(index)
    return self._features.encode_example(decoded_example)


class SerializedRandomFakeGenerator(RandomFakeGenerator):
  """Generator of fake serialized examples."""

  def __iter__(self):
    """Yields all fake examples."""
    for ex in super().__iter__():
      yield self._features.serialize_example(ex)

  def __getitem__(self, index: int):
    example = super().__getitem__(index)
    return self._features.serialize_example(example)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Script to clean up old generated datasets.

The script will display the datasets to delete and will ask users for
confirmation.

Instructions:

python -m tensorflow_datasets.scripts.cleanup.delete_old_versions
    --data_dir=/path/to/data_dir/ --skip_confirmation=False

"""

import collections
import concurrent.futures
import dataclasses
import os
import pathlib
from typing import Dict, Iterable, List, Tuple

from absl import app
from absl import flags
import tensorflow_datasets as tfds
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import termcolor

FLAGS = flags.FLAGS

flags.DEFINE_string(
    'data_dir', tfds.core.constants.DATA_DIR, 'Path to the data directory.'
)
flags.DEFINE_boolean(
    'skip_confirmation', False, 'Whether to skip user confirmation or not.'
)

# Nested dict representing the file structure.
TreeDict = Dict[str, 'TreeDict']  # pytype: disable=not-supported-yet

# Folder in this lists are never deleted
DIRS_TO_KEEP = frozenset({'downloads', 'download', 'manual', 'extracted'})


def _tree_dict() -> TreeDict:
  """Nested dict."""
  return collections.defaultdict(_tree_dict)


def _split_full_names(full_names: List[str]) -> TreeDict:
  """Split the list of full names into a tree of nested dict."""
  root_tree = _tree_dict()
  for full_name in full_names:
    tree = root_tree
    for part in full_name.split('/'):
      tree = tree[part]  # `ds/config/version` -> tree[ds][config][version]
  return root_tree


@dataclasses.dataclass
class TaskResult:
  """Structure containing the tasks result.

  Attributes:
    to_keep: List of directories to keep
    to_delete: List of directories to delete
    sub_tasks: Childs tasks
  """

  to_keep: Iterable[pathlib.Path]
  to_delete: Iterable[pathlib.Path]
  sub_tasks: Iterable[concurrent.futures.Future]


def _extract_dirs_to_delete(
    executor: concurrent.futures.ThreadPoolExecutor,
    curr_dir: pathlib.Path,
    curr_tree: TreeDict,
) -> TaskResult:
  """Tasks which compute the directories to keep and delete.

  This recursively computes `listdir()` on the `curr_dir`, and compares with
  `curr_tree` to check which directory to delete. Recursion is done on the
  remaining sub-directory (cleanup `data_dir/...`, then `data_dir/mnist/...`,
  ...)

  Args:
    executor: Executor to launch subtasks
    curr_dir: Current directory to explore
    curr_tree: Current tree containing the name of folders to keep

  Returns:
    The dataclass containing files to keep, delete and sub-tasks
  """
  # End recursion with tree leafs (`.../1.2.0/`)
  if not curr_tree:
    return TaskResult(
        to_keep=[curr_dir],
        to_delete=[],
        sub_tasks=[],
    )

  # Compute the list of dirs to keep and delete.
  all_files = set(tf.io.gfile.listdir(str(curr_dir)))
  all_dirs_to_keep = set(curr_tree)
  dirs_to_delete = all_files - all_dirs_to_keep - DIRS_TO_KEEP
  dirs_to_maybe_keep = all_files & all_dirs_to_keep

  # Recurse on sub-dirs (e.g. `ds/config/`), add the dirs to the task queue.
  sub_tasks = [
      executor.submit(  # pylint: disable=g-complex-comprehension
          _extract_dirs_to_delete,
          executor=executor,
          curr_dir=curr_dir / dir_name,
          curr_tree=curr_tree[dir_name],
      )
      for dir_name in dirs_to_maybe_keep
  ]
  return TaskResult(
      to_keep=[],
      to_delete=[curr_dir / d for d in dirs_to_delete],
      sub_tasks=sub_tasks,
  )


def _extract_dirs_from_future(
    dirs_to_keep: List[pathlib.Path],
    dirs_to_delete: List[pathlib.Path],
    future: concurrent.futures.Future,
) -> None:
  """Merge all results from all subtasks."""
  task_result = future.result()
  dirs_to_keep.extend(task_result.to_keep)
  dirs_to_delete.extend(task_result.to_delete)
  for sub_task in task_result.sub_tasks:
    _extract_dirs_from_future(
        dirs_to_keep=dirs_to_keep,
        dirs_to_delete=dirs_to_delete,
        future=sub_task,
    )


def _get_extra_dirs(
    data_dir: pathlib.Path,
    current_full_names: List[str],
) -> Tuple[List[pathlib.Path], List[pathlib.Path]]:
  """Returns a tuple of installed datasets and extra dirs to delete.

  Args: data_dir : The path to the data directory current_full_names : Names of
  the latest datasets supported in TFDS

  Returns:
    dirs_to_keep: List of directory to keep (existing supported versions)
    dirs_to_delete: List of directory to delete
  """
  # Normalize paths for windows
  current_full_names = [os.path.normpath(name) for name in current_full_names]

  full_names_tree = _split_full_names(current_full_names)

  # Parse all existing dirs.
  with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
    future = executor.submit(
        _extract_dirs_to_delete,
        executor=executor,
        curr_dir=data_dir,
        curr_tree=full_names_tree,
    )

    dirs_to_keep = []
    dirs_to_delete = []
    _extract_dirs_from_future(
        dirs_to_keep=dirs_to_keep, dirs_to_delete=dirs_to_delete, future=future
    )
  return sorted(dirs_to_keep), sorted(dirs_to_delete)


def _display_dirs(
    data_dir: pathlib.Path,
    dirs_to_keep: List[pathlib.Path],
    dirs_to_delete: List[pathlib.Path],
) -> None:
  """Display dirs to keep and delete."""
  dirs_to_keep = set(dirs_to_keep)
  dirs_to_delete = set(dirs_to_delete)

  # Format dirs to strip prefix and keep in bold.
  def _format(d, keep: bool):
    d = str(d.relative_to(data_dir))
    if keep:
      d = termcolor.colored(f'{d} (*)', attrs=['bold'])
    return d

  all_dirs = [
      _format(d, keep=d in dirs_to_keep)
      for d in sorted(dirs_to_keep | dirs_to_delete)
  ]
  # Preview the dirs to delete
  print(f'The script will delete the following modifications to `{data_dir}`:')
  print('Path indicated in bold (*) will be kept, the other will be deleted.\n')
  print('\n'.join(all_dirs))
  print()
  print(
      f'{len(dirs_to_delete)} datasets to delete, {len(dirs_to_keep)} to '
      'keep.\n'
  )


def delete_old_versions(
    data_dir: pathlib.Path,
    skip_confirmation: bool = False,
) -> None:
  """Detects and deletes all datasets which are not in supported version."""
  # Extract the list of directory to keep and delete.
  dirs_to_keep, dirs_to_delete = _get_extra_dirs(
      data_dir=data_dir,
      # Explicitly set predicate_fn to None to load all datasets
      current_full_names=tfds.core.load.list_full_names(),
  )

  _display_dirs(
      data_dir=data_dir,
      dirs_to_keep=dirs_to_keep,
      dirs_to_delete=dirs_to_delete,
  )

  if skip_confirmation:
    delete = True
  else:
    delete = str(input('Do you want to continue (Y/n): ')) in ('Y', 'y')
  if not delete:
    print('Exiting the program without change!')
    return

  for path in dirs_to_delete:
    print(f'Delete: {path}')
    tf.io.gfile.rmtree(str(path))
  print('Add directories have been deleted!')


def main(_):
  tfds.core.visibility.set_availables([
      tfds.core.visibility.DatasetType.TFDS_PUBLIC,
  ])
  delete_old_versions(
      data_dir=pathlib.Path(FLAGS.data_dir),
      skip_confirmation=FLAGS.skip_confirmation,
  )


if __name__ == '__main__':
  app.run(main)
